"""Basic core types and utilities."""
import os
import time
import functools
import pathlib
import dataclasses

from . import LOCAL_FS_ENCODING
from .utils.log import getLogger
log = getLogger(__name__)

# Audio type selector for no audio.
AUDIO_NONE = 0
# Audio type selector for MPEG (mp3) audio.
AUDIO_MP3 = 1


AUDIO_TYPES = (AUDIO_NONE, AUDIO_MP3)

LP_TYPE = "lp"
EP_TYPE = "ep"
EP_MAX_SIZE_HINT = 6
COMP_TYPE = "compilation"
LIVE_TYPE = "live"
VARIOUS_TYPE = "various"
DEMO_TYPE = "demo"
SINGLE_TYPE = "single"
ALBUM_TYPE_IDS = [LP_TYPE, EP_TYPE, COMP_TYPE, LIVE_TYPE, VARIOUS_TYPE,
                  DEMO_TYPE, SINGLE_TYPE]

VARIOUS_ARTISTS = "Various Artists"

# A key that can be used in a TXXX frame to specify the type of collection
# (or album) a file belongs. See :class:`eyed3.core.ALBUM_TYPE_IDS`.
TXXX_ALBUM_TYPE = "eyeD3#album_type"

# A key that can be used in a TXXX frame to specify the origin of an
# artist/band. i.e. where they are from.
# The format is: city<tab>state<tab>country
TXXX_ARTIST_ORIGIN = "eyeD3#artist_origin"


@dataclasses.dataclass
class ArtistOrigin:
    city: str
    state: str
    country: str

    def __bool__(self):
        return bool(self.city or self.state or self.country)

    def id3Encode(self):
        return "\t".join([(o if o else "") for o in dataclasses.astuple(self)])


class AudioInfo:
    """A base container for common audio details."""

    # The number of seconds of audio data (i.e., the playtime)
    time_secs = 0.0
    # The number of bytes of audio data.
    size_bytes = 0


class Tag:
    """An abstract interface for audio tag (meta) data (e.g. artist, title,
    etc.)
    """

    read_only = False

    def _setArtist(self, val):
        raise NotImplementedError()  # pragma: nocover

    def _getArtist(self):
        raise NotImplementedError()  # pragma: nocover

    def _getAlbumArtist(self):
        raise NotImplementedError()  # pragma: nocover

    def _setAlbumArtist(self, val):
        raise NotImplementedError()  # pragma: nocover

    def _setAlbum(self, val):
        raise NotImplementedError()  # pragma: nocover

    def _getAlbum(self):
        raise NotImplementedError()  # pragma: nocover

    def _setTitle(self, val):
        raise NotImplementedError()  # pragma: nocover

    def _getTitle(self):
        raise NotImplementedError()  # pragma: nocover

    def _setTrackNum(self, val):
        raise NotImplementedError()  # pragma: nocover

    def _getTrackNum(self):
        raise NotImplementedError()  # pragma: nocover

    @property
    def artist(self):
        return self._getArtist()

    @artist.setter
    def artist(self, v):
        self._setArtist(v)

    @property
    def album_artist(self):
        return self._getAlbumArtist()

    @album_artist.setter
    def album_artist(self, v):
        self._setAlbumArtist(v)

    @property
    def album(self):
        return self._getAlbum()

    @album.setter
    def album(self, v):
        self._setAlbum(v)

    @property
    def title(self):
        return self._getTitle()

    @title.setter
    def title(self, v):
        self._setTitle(v)

    @property
    def track_num(self):
        """Track number property.
        Must return a 2-tuple of (track-number, total-number-of-tracks).
        Either tuple value may be ``None``.
        """
        return self._getTrackNum()

    @track_num.setter
    def track_num(self, v):
        self._setTrackNum(v)

    def __init__(self, title=None, artist=None, album=None, album_artist=None, track_num=None):
        self.title = title
        self.artist = artist
        self.album = album
        self.album_artist = album_artist
        self.track_num = track_num


class AudioFile:
    """Abstract base class for audio file types (AudioInfo + Tag)"""

    def _read(self):
        """Subclasses MUST override this method and set ``self._info``,
        ``self._tag`` and ``self.type``.
        """
        raise NotImplementedError()

    def initTag(self, version=None):
        raise NotImplementedError()

    def rename(self, name, fsencoding=LOCAL_FS_ENCODING,
               preserve_file_time=False):
        """Rename the file to ``name``.
        The encoding used for the file name is :attr:`eyed3.LOCAL_FS_ENCODING`
        unless overridden by ``fsencoding``. Note, if the target file already
        exists, or the full path contains non-existent directories the
        operation will fail with :class:`IOError`.
        File times are not modified when ``preserve_file_time`` is ``True``,
        ``False`` is the default.
        """
        curr_path = pathlib.Path(self.path)
        ext = curr_path.suffix

        new_path = curr_path.parent / "{name}{ext}".format(**locals())
        if new_path.exists():
            raise IOError("File '%s' exists, will not overwrite" % new_path)
        elif not new_path.parent.exists():
            raise IOError("Target directory '%s' does not exists, will not "
                          "create" % new_path.parent)

        os.rename(self.path, str(new_path))
        if self.tag:
            self.tag.file_info.name = str(new_path)
        if preserve_file_time:
            self.tag.file_info.touch((self.tag.file_info.atime,
                                      self.tag.file_info.mtime))

        self.path = str(new_path)

    @property
    def path(self):
        """The absolute path of this file."""
        return self._path

    @path.setter
    def path(self, path):
        """Set the path"""
        if isinstance(path, pathlib.Path):
            path = str(path)
        self._path = path

    @property
    def info(self):
        """Returns a concrete implemenation of :class:`eyed3.core.AudioInfo`"""
        return self._info

    @property
    def tag(self):
        """Returns a concrete implemenation of :class:`eyed3.core.Tag`"""
        return self._tag

    @tag.setter
    def tag(self, t):
        self._tag = t

    def __init__(self, path):
        """Construct with a path and invoke ``_read``.
        All other members are set to None."""
        if isinstance(path, pathlib.Path):
            path = str(path)
        self.path = path

        self.type = None
        self._info = None
        self._tag = None
        self._read()


@functools.total_ordering
class Date:
    """
    A class for representing a date and time (optional). This class differs
    from ``datetime.datetime`` in that the default values for month, day,
    hour, minute, and second is ``None`` and not 'January 1, 00:00:00'.
    This allows for an object that is simply 1987, and not January 1 12AM,
    for example. But when more resolution is required those vales can be set
    as well.
    """

    TIME_STAMP_FORMATS = ["%Y",
                          "%Y-%m",
                          "%Y-%m-%d",
                          "%Y-%m-%dT%H",
                          "%Y-%m-%dT%H:%M",
                          "%Y-%m-%dT%H:%M:%S",
                          # The following end with 'Z' signally time is UTC
                          "%Y-%m-%dT%HZ",
                          "%Y-%m-%dT%H:%MZ",
                          "%Y-%m-%dT%H:%M:%SZ",
                          # The following are wrong per the specs, but ...
                          "%Y-%m-%d %H:%M:%S",
                          "%Y-00-00",
                          "%Y%m%d",
                          ]
    """Valid time stamp formats per ISO 8601 and used by `strptime`."""

    def __init__(self, year, month=None, day=None,
                 hour=None, minute=None, second=None):
        # Validate with datetime
        from datetime import datetime
        _ = datetime(year, month if month is not None else 1,
                     day if day is not None else 1,
                     hour if hour is not None else 0,
                     minute if minute is not None else 0,
                     second if second is not None else 0)

        self._year = year
        self._month = month
        self._day = day
        self._hour = hour
        self._minute = minute
        self._second = second

        # Python's date classes do a lot more date validation than does not
        # need to be duplicated here.  Validate it
        _ = Date._validateFormat(str(self))                           # noqa

    @property
    def year(self):
        return self._year

    @property
    def month(self):
        return self._month

    @property
    def day(self):
        return self._day

    @property
    def hour(self):
        return self._hour

    @property
    def minute(self):
        return self._minute

    @property
    def second(self):
        return self._second

    def __eq__(self, rhs):
        if not rhs:
            return False

        return (self.year == rhs.year and
                self.month == rhs.month and
                self.day == rhs.day and
                self.hour == rhs.hour and
                self.minute == rhs.minute and
                self.second == rhs.second)

    def __ne__(self, rhs):
        return not(self == rhs)

    def __lt__(self, rhs):
        if not rhs:
            return False

        for left, right in ((self.year, rhs.year),
                            (self.month, rhs.month),
                            (self.day, rhs.day),
                            (self.hour, rhs.hour),
                            (self.minute, rhs.minute),
                            (self.second, rhs.second)):

            left = left if left is not None else -1
            right = right if right is not None else -1

            if left < right:
                return True
            elif left > right:
                return False

        return False

    def __hash__(self):
        return hash(str(self))

    @staticmethod
    def _validateFormat(s):
        pdate, fmt = None, None
        for fmt in Date.TIME_STAMP_FORMATS:
            try:
                pdate = time.strptime(s, fmt)
                break
            except ValueError:
                # date string did not match format.
                continue

        if pdate is None:
            raise ValueError(f"Invalid date string: {s}")

        assert pdate
        return pdate, fmt

    @staticmethod
    def parse(s):
        """Parses date strings that conform to ISO-8601."""
        if not isinstance(s, str):
            s = s.decode("ascii")
        s = s.strip('\x00')

        pdate, fmt = Date._validateFormat(s)

        # Here is the difference with Python date/datetime objects, some
        # of the members can be None
        kwargs = {}
        if "%m" in fmt:
            kwargs["month"] = pdate.tm_mon
        if "%d" in fmt:
            kwargs["day"] = pdate.tm_mday
        if "%H" in fmt:
            kwargs["hour"] = pdate.tm_hour
        if "%M" in fmt:
            kwargs["minute"] = pdate.tm_min
        if "%S" in fmt:
            kwargs["second"] = pdate.tm_sec

        return Date(pdate.tm_year, **kwargs)

    def __str__(self):
        """Returns date strings that conform to ISO-8601.
        The returned string will be no larger than 17 characters."""
        s = "%d" % self.year
        if self.month:
            s += "-%s" % str(self.month).rjust(2, '0')
            if self.day:
                s += "-%s" % str(self.day).rjust(2, '0')
                if self.hour is not None:
                    s += "T%s" % str(self.hour).rjust(2, '0')
                    if self.minute is not None:
                        s += ":%s" % str(self.minute).rjust(2, '0')
                        if self.second is not None:
                            s += ":%s" % str(self.second).rjust(2, '0')
        return s


def parseError(ex):
    """A function that is invoked when non-fatal parse, format, etc. errors
    occur. In most cases the invalid values will be ignored or possibly fixed.
    This function simply logs the error."""
    log.warning(ex)


def load(path, tag_version=None) -> AudioFile:
    """Loads the file identified by ``path`` and returns a concrete type of
    :class:`eyed3.core.AudioFile`. If ``path`` is not a file an ``IOError`` is
    raised. ``None`` is returned when the file type (i.e. mime-type) is not
    recognized.
    The following AudioFile types are supported:

      * :class:`eyed3.mp3.Mp3AudioFile` - For mp3 audio files.
      * :class:`eyed3.id3.TagFile` - For raw ID3 data files.

    If ``tag_version`` is not None (the default) only a specific version of
    metadata is loaded. This value must be a version constant specific to the
    eventual format of the metadata.
    """
    from . import mp3, id3
    from .mimetype import guessMimetype

    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)
    log.debug(f"Loading file: {path}")

    if path.exists():
        if not path.is_file():
            raise IOError(f"not a file: {path}")
    else:
        raise IOError(f"file not found: {path}")

    mtype = guessMimetype(path)
    log.debug(f"File mime-type: {mtype}")

    if mtype in mp3.MIME_TYPES:
        return mp3.Mp3AudioFile(path, tag_version)
    elif mtype == id3.ID3_MIME_TYPE:
        return id3.TagFile(path, tag_version)
    else:
        return None



import time
import pprint
import eyed3
import eyed3.utils
from pathlib import Path
from collections import Counter
from eyed3.mimetype import guessMimetype
from eyed3.utils.log import getLogger

log = getLogger(__name__)

# python-magic
try:
    import magic

    class MagicTypes(magic.Magic):
        def __init__(self):
            magic.Magic.__init__(self, mime=True, mime_encoding=False, keep_going=True)

        def guess_type(self, filename, all_types=False):
            try:
                types = self.from_file(filename)
            except UnicodeEncodeError:
                # https://github.com/ahupp/python-magic/pull/144
                types = self.from_file(filename.encode("utf-8", 'surrogateescape'))

            delim = r"\012- "
            if all_types:
                return types.split(delim)
            else:
                return types.split(delim)[0]

    _python_magic = MagicTypes()

except ImportError:
    _python_magic = None


class MimetypesPlugin(eyed3.plugins.LoaderPlugin):
    NAMES = ["mimetypes"]

    def __init__(self, arg_parser):
        self._num_visited = 0
        super().__init__(arg_parser, cache_files=False, track_images=False)

        g = self.arg_group
        g.add_argument("--status", action="store_true", help="Print dot status.")
        g.add_argument("--parse-files", action="store_true", help="Parse each file.")
        g.add_argument("--hide-notfound", action="store_true")
        if _python_magic:
            g.add_argument("-M", "--use-pymagic", action="store_true",
                           help="Use python-magic to determine mimetype.")
        self.magic = None
        self.start_t = None
        self.mime_types = Counter()

    def start(self, args, config):
        super().start(args, config)
        self.magic = "pymagic" if self.args.use_pymagic else "filetype"
        self.start_t = time.time()

    def handleFile(self, f, *args, **kwargs):

        self._num_visited += 1
        if self.args.parse_files:
            try:
                super().handleFile(f)
            except Exception as ex:
                log.critical(ex, exc_info=ex)
        else:
            self._num_loaded += 1

            if self.magic == "pymagic":
                mtype = _python_magic.guess_type(f)
            else:
                mtype = guessMimetype(f)

            self.mime_types[mtype] += 1
            if not self.args.hide_notfound:
                if mtype is None and Path(f).suffix.lower() in (".mp3",):
                    print("None mimetype:", f)

        if self.args.status:
            print(".", end="", flush=True)

    def handleDone(self):
        t = time.time() - self.start_t
        print(f"\nVisited {self._num_visited} files")
        print(f"Processed {self._num_loaded} files")
        print(f"magic: {self.magic}")
        print(f"time: {eyed3.utils.formatTime(t)} seconds")
        if self.mime_types:
            pprint.pprint(self.mime_types)



from .__regarding__ import *  # noqa: F403

__project_name__ = project_name
__version__ = version
__version_info__ = version_info
__release__ = version_info.release
__release_name__ = release_name
__years__ = years

__project_slug__ = "eyed3"
__pypi_name__ = "eyeD3"
__author__ = author
__author_email__ = author_email
__url__ = homepage
__description__ = description
# FIXME: __long_description__ not being used anywhere.
__long_description__ = """
eyeD3 is a Python module and command line program for processing ID3 tags.
Information about mp3 files (i.e bit rate, sample frequency,
play time, etc.) is also provided. The formats supported are ID3
v1.0/v1.1 and v2.3/v2.4.
"""
__license__ = "GNU GPL v3.0"
__github_url__ = "https://github.com/nicfit/eyeD3",
__version_txt__ = """%(__project_name__)s %(__version__)s © Copyright %(__years__)s %(__author__)s
This program comes with ABSOLUTELY NO WARRANTY! See LICENSE for details.
Run with --help/-h for usage information or read the docs at
%(__url__)s""" % (locals())



from eyed3.plugins import LoaderPlugin
from eyed3.id3.apple import PCST, PCST_FID, WFED, WFED_FID


class Podcast(LoaderPlugin):
    NAMES = ['itunes-podcast']
    SUMMARY = "Adds (or removes) the tags necessary for Apple iTunes to "\
              "identify the file as a podcast."

    def __init__(self, arg_parser):
        super(Podcast, self).__init__(arg_parser)
        g = self.arg_group
        g.add_argument("--add", action="store_true",
                       help="Add the podcast frames.")
        g.add_argument("--remove", action="store_true",
                       help="Remove the podcast frames.")

    def _add(self, tag):
        save = False
        if PCST_FID not in tag.frame_set:
            tag.frame_set[PCST_FID] = PCST()
            save = True
        if WFED_FID not in tag.frame_set:
            tag.frame_set[WFED_FID] = WFED("http://eyeD3.nicfit.net/")
            save = True

        if save:
            print("\tAdding...")
            tag.save(backup=self.args.backup)
            self._printStatus(tag)

    def _remove(self, tag):
        save = False
        for fid in [PCST_FID, WFED_FID]:
            try:
                del tag.frame_set[fid]
                save = True
            except KeyError:
                continue

        if save:
            print("\tRemoving...")
            tag.save(backup=self.args.backup)
            self._printStatus(tag)

    def _printStatus(self, tag):
        status = ":-("
        if PCST_FID in tag.frame_set:
            status = ":-/"
            if WFED_FID in tag.frame_set:
                status = ":-)"
        print("\tiTunes podcast? %s" % status)

    def handleFile(self, f):
        super(Podcast, self).handleFile(f)

        if self.audio_file and self.audio_file.tag:
            print(f)
            tag = self.audio_file.tag
            self._printStatus(tag)
            if self.args.remove:
                self._remove(self.audio_file.tag)
            elif self.args.add:
                self._add(self.audio_file.tag)



import os
import re

from .. import Error
from .. import id3
from .. import core

from ..utils.log import getLogger
log = getLogger(__name__)


class Mp3Exception(Error):
    """Used to signal mp3-related errors."""
    pass


NAME = "mpeg"
# Mime-types that are recognized at MP3
MIME_TYPES = ["audio/mpeg", "audio/mp3", "audio/x-mp3", "audio/x-mpeg",
              "audio/mpeg3", "audio/x-mpeg3", "audio/mpg", "audio/x-mpg",
              "audio/x-mpegaudio", "audio/mpegapplication/x-tar",
             ]

# Mime-types that have been seen to contain mp3 data.
OTHER_MIME_TYPES = ['application/octet-stream',  # ???
                    'audio/x-hx-aac-adts',  # ???
                    'audio/x-wav',  # RIFF wrapped mp3s
                   ]

# Valid file extensions.
EXTENSIONS = [".mp3"]


class Mp3AudioInfo(core.AudioInfo):
    def __init__(self, file_obj, start_offset, tag):
        from . import headers
        from .headers import timePerFrame

        log.debug("mp3 header search starting @ %x" % start_offset)
        core.AudioInfo.__init__(self)

        self.mp3_header = None
        self.xing_header = None
        self.vbri_header = None
        # If not ``None``, the Lame header.
        # See :class:`eyed3.mp3.headers.LameHeader`
        self.lame_tag = None
        # 2-tuple, (vrb?:boolean, bitrate:int)
        self.bit_rate = (None, None)

        header_pos = 0
        while self.mp3_header is None:
            # Find first mp3 header
            (header_pos,
             header_int,
             header_bytes) = headers.findHeader(file_obj, start_offset)
            if not header_int:
                try:
                    fname = file_obj.name
                except AttributeError:
                    fname = 'unknown'
                raise headers.Mp3Exception(
                    "Unable to find a valid mp3 frame in '%s'" % fname)

            try:
                self.mp3_header = headers.Mp3Header(header_int)
                log.debug("mp3 header %x found at position: 0x%x" %
                          (header_int, header_pos))
            except headers.Mp3Exception as ex:
                log.debug("Invalid mp3 header: %s" % str(ex))
                # keep looking...
                start_offset += 4

        file_obj.seek(header_pos)
        mp3_frame = file_obj.read(self.mp3_header.frame_length)
        if re.compile(b'Xing|Info').search(mp3_frame):
            # Check for Xing/Info header information.
            self.xing_header = headers.XingHeader()
            if not self.xing_header.decode(mp3_frame):
                log.debug("Ignoring corrupt Xing header")
                self.xing_header = None
        elif mp3_frame.find(b'VBRI') >= 0:
            # Check for VBRI header information.
            self.vbri_header = headers.VbriHeader()
            if not self.vbri_header.decode(mp3_frame):
                log.debug("Ignoring corrupt VBRI header")
                self.vbri_header = None

        # Check for LAME Tag
        self.lame_tag = headers.LameHeader(mp3_frame)

        # Set file size
        import stat
        self.size_bytes = os.stat(file_obj.name)[stat.ST_SIZE]

        # Compute track play time.
        if self.xing_header and self.xing_header.vbr:
            tpf = timePerFrame(self.mp3_header, True)
            self.time_secs = tpf * self.xing_header.numFrames
        elif self.vbri_header and self.vbri_header.version == 1:
            tpf = timePerFrame(self.mp3_header, True)
            self.time_secs = tpf * self.vbri_header.num_frames
        else:
            tpf = timePerFrame(self.mp3_header, False)
            length = self.size_bytes
            if tag and tag.isV2():
                length -= tag.header.SIZE + tag.header.tag_size
                # Handle the case where there is a v2 tag and a v1 tag.
                file_obj.seek(-128, 2)
                if file_obj.read(3) == "TAG":
                    length -= 128
            elif tag and tag.isV1():
                length -= 128
            self.time_secs = (length / self.mp3_header.frame_length) * tpf

        # Compute bitrate
        if (self.xing_header and self.xing_header.vbr and
                self.xing_header.numFrames):  # if xing_header.numFrames == 0, ZeroDivisionError
            br = int((self.xing_header.numBytes * 8) /
                     (tpf * self.xing_header.numFrames * 1000))
            vbr = True
        else:
            br = self.mp3_header.bit_rate
            vbr = False
        self.bit_rate = (vbr, br)

        self.sample_freq = self.mp3_header.sample_freq
        self.mode = self.mp3_header.mode

    ##
    # Helper to get the bitrate as a string. The prefix '~' is used to denote
    # variable bit rates.
    @property
    def bit_rate_str(self):
        (vbr, bit_rate) = self.bit_rate
        return f"{'~' if vbr else ''}{bit_rate} kb/s"


class Mp3AudioFile(core.AudioFile):
    """Audio file container for mp3 files."""

    def __init__(self, path, version=id3.ID3_ANY_VERSION):
        self._tag_version = version

        super().__init__(path)
        assert self.type == core.AUDIO_MP3

    def _read(self):
        with open(self.path, "rb") as file_obj:
            self._tag = id3.Tag()
            tag_found = self._tag.parse(file_obj, self._tag_version)

            # Compute offset for starting mp3 data search
            if tag_found and self._tag.isV1():
                mp3_offset = 0
            elif tag_found and self._tag.isV2():
                mp3_offset = self._tag.header.SIZE + self._tag.header.tag_size
            else:
                mp3_offset = 0
                self._tag = None

            try:
                self._info = Mp3AudioInfo(file_obj, mp3_offset, self._tag)
            except Mp3Exception as ex:
                # Only logging a warning here since we can still operate on
                # the tag.
                log.warning(ex)
                self._info = None

            self.type = core.AUDIO_MP3

    def initTag(self, version=id3.ID3_DEFAULT_VERSION):
        """Add a id3.Tag to the file (removing any existing tag if one exists)."""
        self.tag = id3.Tag()
        self.tag.version = version
        self.tag.file_info = id3.FileInfo(self.path)
        return self.tag

    @core.AudioFile.tag.setter
    def tag(self, t):
        if t:
            t.file_info = id3.FileInfo(self.path)
            if self._tag and self._tag.file_info:
                t.file_info.tag_size = self._tag.file_info.tag_size
                t.file_info.tag_padding_size = \
                    self._tag.file_info.tag_padding_size
        self._tag = t



import os
import re
import dataclasses
from functools import partial
from argparse import ArgumentTypeError

from eyed3.plugins import LoaderPlugin
from eyed3 import core, id3, mp3
from eyed3.utils import makeUniqueFileName, b, formatTime
from eyed3.utils.console import (
    printMsg, printError, printWarning, boldText, getTtySize,
)
from eyed3.id3.frames import ImageFrame
from eyed3.mimetype import guessMimetype

from eyed3.utils.log import getLogger
log = getLogger(__name__)

FIELD_DELIM = ':'

DEFAULT_MAX_PADDING = 64 * 1024


class ClassicPlugin(LoaderPlugin):
    SUMMARY = "Classic eyeD3 interface for viewing and editing tags."
    DESCRIPTION = """
All PATH arguments are parsed and displayed. Directory paths are searched
recursively. Any editing options (--artist, --title) are applied to each file
read.

All date options (-Y, --release-year excepted) follow ISO 8601 format. This is
``yyyy-mm-ddThh:mm:ss``. The year is required, and each component thereafter is
optional. For example, 2012-03 is valid, 2012--12 is not.
"""
    NAMES = ["classic"]

    def __init__(self, arg_parser):
        super(ClassicPlugin, self).__init__(arg_parser)
        g = self.arg_group

        def PositiveIntArg(i):
            i = int(i)
            if i < 0:
                raise ArgumentTypeError("positive number required")
            return i

        # Common options
        g.add_argument("-a", "--artist", dest="artist",
                       metavar="STRING", help=ARGS_HELP["--artist"])
        g.add_argument("-A", "--album", dest="album",
                       metavar="STRING", help=ARGS_HELP["--album"])
        g.add_argument("-b", "--album-artist",
                       dest="album_artist", metavar="STRING",
                       help=ARGS_HELP["--album-artist"])
        g.add_argument("-t", "--title", dest="title",
                       metavar="STRING", help=ARGS_HELP["--title"])
        g.add_argument("-n", "--track", type=PositiveIntArg, dest="track",
                       metavar="NUM", help=ARGS_HELP["--track"])
        g.add_argument("-N", "--track-total", type=PositiveIntArg,
                       dest="track_total", metavar="NUM",
                       help=ARGS_HELP["--track-total"])

        g.add_argument("--track-offset", type=int, dest="track_offset",
                       metavar="N", help=ARGS_HELP["--track-offset"])

        g.add_argument("--composer", dest="composer",
                       metavar="STRING", help=ARGS_HELP["--composer"])
        g.add_argument("--orig-artist", dest="orig_artist",
                       metavar="STRING", help=ARGS_HELP["--orig-artist"])
        g.add_argument("-d", "--disc-num", type=PositiveIntArg, dest="disc_num",
                       metavar="NUM", help=ARGS_HELP["--disc-num"])
        g.add_argument("-D", "--disc-total", type=PositiveIntArg,
                       dest="disc_total", metavar="NUM",
                       help=ARGS_HELP["--disc-total"])
        g.add_argument("-G", "--genre", dest="genre",
                       metavar="GENRE", help=ARGS_HELP["--genre"])
        g.add_argument("--non-std-genres", dest="non_std_genres",
                       action="store_true", help=ARGS_HELP["--non-std-genres"])
        g.add_argument("-Y", "--release-year", type=PositiveIntArg,
                       dest="release_year", metavar="YEAR",
                       help=ARGS_HELP["--release-year"])
        g.add_argument("-c", "--comment", dest="simple_comment",
                       metavar="STRING",
                       help=ARGS_HELP["--comment"])
        g.add_argument("--artist-city", metavar="STRING",
                       help="The artist's city of origin. "
                            f"Stored as a user text frame `{core.TXXX_ARTIST_ORIGIN}`")
        g.add_argument("--artist-state", metavar="STRING",
                       help="The artist's state of origin. "
                       f"Stored as a user text frame `{core.TXXX_ARTIST_ORIGIN}`")
        g.add_argument("--artist-country", metavar="STRING",
                       help="The artist's country of origin. "
                       f"Stored as a user text frame `{core.TXXX_ARTIST_ORIGIN}`")
        g.add_argument("--rename", dest="rename_pattern", metavar="PATTERN",
                       help=ARGS_HELP["--rename"])

        gid3 = arg_parser.add_argument_group("ID3 options")

        def _splitArgs(arg, maxsplit=None):
            NEW_DELIM = "#DELIM#"
            arg = re.sub(r"\\%s" % FIELD_DELIM, NEW_DELIM, arg)
            t = tuple(re.sub(NEW_DELIM, FIELD_DELIM, s)
                         for s in arg.split(FIELD_DELIM))
            if maxsplit is not None and maxsplit < 2:
                raise ValueError("Invalid maxsplit value: {}".format(maxsplit))
            elif maxsplit and len(t) > maxsplit:
                t = t[:maxsplit - 1] + (FIELD_DELIM.join(t[maxsplit - 1:]),)
                assert len(t) <= maxsplit
            return t

        def DescLangArg(arg):
            """DESCRIPTION[:LANG]"""
            vals = _splitArgs(arg, 2)
            desc = vals[0]
            lang = vals[1] if len(vals) > 1 else id3.DEFAULT_LANG
            return desc, b(lang)[:3] or id3.DEFAULT_LANG

        def DescTextArg(arg):
            """DESCRIPTION:TEXT"""
            vals = _splitArgs(arg, 2)
            desc = vals[0].strip()
            text = FIELD_DELIM.join(vals[1:] if len(vals) > 1 else [])
            return desc or "", text or ""
        KeyValueArg = DescTextArg

        def DescUrlArg(arg):
            desc, url = DescTextArg(arg)
            return desc, url.encode("latin1")

        def FidArg(arg):
            fid = arg.strip().encode("ascii")
            if not fid:
                raise ArgumentTypeError("No frame ID")
            return fid

        def TextFrameArg(arg):
            """FID:TEXT"""
            vals = _splitArgs(arg, 2)
            fid = vals[0].strip().encode("ascii")
            if not fid:
                raise ArgumentTypeError("No frame ID")
            text = vals[1] if len(vals) > 1 else ""
            return fid, text

        def UrlFrameArg(arg):
            """FID:TEXT"""
            fid, url = TextFrameArg(arg)
            return fid, url.encode("latin1")

        def DateArg(date_str):
            return core.Date.parse(date_str) if date_str else ""

        def CommentArg(arg):
            """
            COMMENT[:DESCRIPTION[:LANG]
            """
            vals = _splitArgs(arg, 3)
            text = vals[0]
            if not text:
                raise ArgumentTypeError("text required")
            desc = vals[1] if len(vals) > 1 else ""
            lang = vals[2] if len(vals) > 2 else id3.DEFAULT_LANG
            return text, desc, b(lang)[:3]

        def LyricsArg(arg):
            text, desc, lang = CommentArg(arg)
            try:
                with open(text, "r") as fp:
                    data = fp.read()
            except Exception:                                       # noqa: B901
                raise ArgumentTypeError("Unable to read file")
            return data, desc, lang

        def PlayCountArg(pc):
            if not pc:
                raise ArgumentTypeError("value required")
            increment = False
            if pc[0] == "+":
                pc = int(pc[1:])
                increment = True
            else:
                pc = int(pc)
            if pc < 0:
                raise ArgumentTypeError("out of range")
            return increment, pc

        def BpmArg(bpm):
            bpm = int(float(bpm) + 0.5)
            if bpm <= 0:
                raise ArgumentTypeError("out of range")
            return bpm

        def DirArg(d):
            if not d or not os.path.isdir(d):
                raise ArgumentTypeError("invalid directory: %s" % d)
            return d

        def ImageArg(s):
            """PATH:TYPE[:DESCRIPTION]
            Returns (path, type_id, mime_type, description)
            """
            args = _splitArgs(s, 3)
            if len(args) < 2:
                raise ArgumentTypeError("Format is: PATH:TYPE[:DESCRIPTION]")

            path, type_str = args[:2]
            desc = args[2] if len(args) > 2 else ""

            try:
                type_id = id3.frames.ImageFrame.stringToPicType(type_str)
            except Exception:                                                 # noqa: B901
                raise ArgumentTypeError("invalid pic type: {}".format(type_str))

            if not path:
                raise ArgumentTypeError("path required")
            elif True in [path.startswith(prefix)
                          for prefix in ["http://", "https://"]]:
                mt = ImageFrame.URL_MIME_TYPE
            else:
                if not os.path.isfile(path):
                    raise ArgumentTypeError("file does not exist")
                mt = guessMimetype(path)
                if mt is None:
                    raise ArgumentTypeError("Cannot determine mime-type")

            return path, type_id, mt, desc

        def ObjectArg(s):
            """OBJ_PATH:MIME-TYPE[:DESCRIPTION[:FILENAME]],
            Returns (path, mime_type, description, filename)
            """
            args = _splitArgs(s, 4)
            if len(args) < 2:
                raise ArgumentTypeError("too few parts")

            path = args[0]
            if path:
                mt = args[1]
                desc = args[2] if len(args) > 2 else ""
                filename = args[3] \
                             if len(args) > 3 \
                                else os.path.basename(path)
                if not os.path.isfile(path):
                    raise ArgumentTypeError("file does not exist")
                if not mt:
                    raise ArgumentTypeError("mime-type required")
            else:
                raise ArgumentTypeError("path required")
            return (path, mt, desc, filename)

        def UniqFileIdArg(arg):
            owner_id, id = KeyValueArg(arg)
            if not owner_id:
                raise ArgumentTypeError("owner_id required")
            id = id.encode("latin1")  # don't want to pass unicode
            if len(id) > 64:
                raise ArgumentTypeError("id must be <= 64 bytes")
            return (owner_id, id)

        def PopularityArg(arg):
            """EMAIL:RATING[:PLAY_COUNT]
            Returns (email, rating, play_count)
            """
            args = _splitArgs(arg, 3)
            if len(args) < 2:
                raise ArgumentTypeError("Incorrect number of argument components")
            email = args[0]
            rating = int(float(args[1]))
            if rating < 0 or rating > 255:
                raise ArgumentTypeError("Rating out-of-range")
            play_count = 0
            if len(args) > 2:
                play_count = int(args[2])
            if play_count < 0:
                raise ArgumentTypeError("Play count out-of-range")
            return (email, rating, play_count)

        # Tag versions
        gid3.add_argument("-1", "--v1", action="store_const", const=id3.ID3_V1,
                          dest="tag_version", default=id3.ID3_ANY_VERSION,
                          help=ARGS_HELP["--v1"])
        gid3.add_argument("-2", "--v2", action="store_const", const=id3.ID3_V2,
                          dest="tag_version", default=id3.ID3_ANY_VERSION,
                          help=ARGS_HELP["--v2"])
        gid3.add_argument("--to-v1.1", action="store_const", const=id3.ID3_V1_1,
                          dest="convert_version", help=ARGS_HELP["--to-v1.1"])
        gid3.add_argument("--to-v2.3", action="store_const", const=id3.ID3_V2_3,
                          dest="convert_version", help=ARGS_HELP["--to-v2.3"])
        gid3.add_argument("--to-v2.4", action="store_const", const=id3.ID3_V2_4,
                          dest="convert_version", help=ARGS_HELP["--to-v2.4"])

        # Dates
        gid3.add_argument("--release-date", type=DateArg, dest="release_date",
                          metavar="DATE",
                          help=ARGS_HELP["--release-date"])
        gid3.add_argument("--orig-release-date", type=DateArg,
                          dest="orig_release_date", metavar="DATE",
                          help=ARGS_HELP["--orig-release-date"])
        gid3.add_argument("--recording-date", type=DateArg,
                          dest="recording_date", metavar="DATE",
                          help=ARGS_HELP["--recording-date"])
        gid3.add_argument("--encoding-date", type=DateArg, dest="encoding_date",
                          metavar="DATE", help=ARGS_HELP["--encoding-date"])
        gid3.add_argument("--tagging-date", type=DateArg, dest="tagging_date",
                          metavar="DATE", help=ARGS_HELP["--tagging-date"])

        # Misc
        gid3.add_argument("--publisher", action="store",
                          dest="publisher", metavar="STRING",
                          help=ARGS_HELP["--publisher"])
        gid3.add_argument("--play-count", type=PlayCountArg, dest="play_count",
                          metavar="<+>N", default=None,
                          help=ARGS_HELP["--play-count"])
        gid3.add_argument("--bpm", type=BpmArg, dest="bpm", metavar="N",
                          default=None, help=ARGS_HELP["--bpm"])
        gid3.add_argument("--unique-file-id", action="append",
                          type=UniqFileIdArg, dest="unique_file_ids",
                          metavar="OWNER_ID:ID", default=[],
                          help=ARGS_HELP["--unique-file-id"])

        # Comments
        gid3.add_argument("--add-comment", action="append", dest="comments",
                          metavar="COMMENT[:DESCRIPTION[:LANG]]", default=[],
                          type=CommentArg, help=ARGS_HELP["--add-comment"])
        gid3.add_argument("--remove-comment", action="append", type=DescLangArg,
                          dest="remove_comment", default=[],
                          metavar="DESCRIPTION[:LANG]",
                          help=ARGS_HELP["--remove-comment"])
        gid3.add_argument("--remove-all-comments", action="store_true",
                          dest="remove_all_comments",
                          help=ARGS_HELP["--remove-all-comments"])

        gid3.add_argument("--add-lyrics", action="append", type=LyricsArg,
                          dest="lyrics", default=[],
                          metavar="LYRICS_FILE[:DESCRIPTION[:LANG]]",
                          help=ARGS_HELP["--add-lyrics"])
        gid3.add_argument("--remove-lyrics", action="append", type=DescLangArg,
                          dest="remove_lyrics", default=[],
                          metavar="DESCRIPTION[:LANG]",
                          help=ARGS_HELP["--remove-lyrics"])
        gid3.add_argument("--remove-all-lyrics", action="store_true",
                          dest="remove_all_lyrics",
                          help=ARGS_HELP["--remove-all-lyrics"])

        gid3.add_argument("--text-frame", action="append", type=TextFrameArg,
                          dest="text_frames", metavar="FID:TEXT", default=[],
                          help=ARGS_HELP["--text-frame"])
        gid3.add_argument("--user-text-frame", action="append",
                          type=DescTextArg,
                          dest="user_text_frames", metavar="DESC:TEXT",
                          default=[], help=ARGS_HELP["--user-text-frame"])

        gid3.add_argument("--url-frame", action="append", type=UrlFrameArg,
                          dest="url_frames", metavar="FID:URL", default=[],
                          help=ARGS_HELP["--url-frame"])
        gid3.add_argument("--user-url-frame", action="append", type=DescUrlArg,
                          dest="user_url_frames", metavar="DESCRIPTION:URL",
                          default=[], help=ARGS_HELP["--user-url-frame"])

        gid3.add_argument("--add-image", action="append", type=ImageArg,
                          dest="images", metavar="IMG_PATH:TYPE[:DESCRIPTION]",
                          default=[], help=ARGS_HELP["--add-image"])
        gid3.add_argument("--remove-image", action="append",
                          dest="remove_image", default=[],
                          metavar="DESCRIPTION",
                          help=ARGS_HELP["--remove-image"])
        gid3.add_argument("--remove-all-images", action="store_true",
                          dest="remove_all_images",
                          help=ARGS_HELP["--remove-all-images"])
        gid3.add_argument("--write-images", dest="write_images_dir",
                          metavar="DIR", type=DirArg,
                          help=ARGS_HELP["--write-images"])

        gid3.add_argument("--add-object", action="append", type=ObjectArg,
                          dest="objects", default=[],
                          metavar="OBJ_PATH:MIME-TYPE[:DESCRIPTION[:FILENAME]]",
                          help=ARGS_HELP["--add-object"])
        gid3.add_argument("--remove-object", action="append",
                          dest="remove_object", default=[],
                          metavar="DESCRIPTION",
                          help=ARGS_HELP["--remove-object"])
        gid3.add_argument("--write-objects", action="store",
                          dest="write_objects_dir", metavar="DIR", default=None,
                          help=ARGS_HELP["--write-objects"])
        gid3.add_argument("--remove-all-objects", action="store_true",
                          dest="remove_all_objects",
                          help=ARGS_HELP["--remove-all-objects"])

        gid3.add_argument("--add-popularity", action="append",
                          type=PopularityArg, dest="popularities", default=[],
                          metavar="EMAIL:RATING[:PLAY_COUNT]",
                          help=ARGS_HELP["--add-popularity"])
        gid3.add_argument("--remove-popularity", action="append", type=str,
                          dest="remove_popularity", default=[],
                          metavar="EMAIL",
                          help=ARGS_HELP["--remove-popularity"])

        gid3.add_argument("--remove-v1", action="store_true", dest="remove_v1",
                          default=False, help=ARGS_HELP["--remove-v1"])
        gid3.add_argument("--remove-v2", action="store_true", dest="remove_v2",
                          default=False, help=ARGS_HELP["--remove-v2"])
        gid3.add_argument("--remove-all", action="store_true", default=False,
                          dest="remove_all", help=ARGS_HELP["--remove-all"])
        gid3.add_argument("--remove-frame", action="append", default=[],
                          dest="remove_fids", metavar="FID", type=FidArg,
                          help=ARGS_HELP["--remove-frame"])

        # 'True' means 'apply default max_padding, but only if saving anyhow'
        gid3.add_argument("--max-padding", type=int, dest="max_padding",
                          default=True, metavar="NUM_BYTES",
                          help=ARGS_HELP["--max-padding"])
        gid3.add_argument("--no-max-padding", dest="max_padding",
                          action="store_const", const=None,
                          help=ARGS_HELP["--no-max-padding"])

        _encodings = ["latin1", "utf8", "utf16", "utf16-be"]
        gid3.add_argument("--encoding", dest="text_encoding", default=None,
                          choices=_encodings, metavar='|'.join(_encodings),
                          help=ARGS_HELP["--encoding"])

        # Misc options
        gid4 = arg_parser.add_argument_group("Misc options")
        gid4.add_argument("--force-update", action="store_true", default=False,
                          dest="force_update", help=ARGS_HELP["--force-update"])
        gid4.add_argument("-v", "--verbose", action="store_true",
                          dest="verbose", help=ARGS_HELP["--verbose"])
        gid4.add_argument("--preserve-file-times", action="store_true",
                          dest="preserve_file_time",
                          help=ARGS_HELP["--preserve-file-times"])

    def handleFile(self, f):
        parse_version = self.args.tag_version

        try:
            super().handleFile(f, tag_version=parse_version)
        except id3.TagException as tag_ex:
            printError(str(tag_ex))
            return

        if not self.audio_file:
            return

        self.terminal_width = getTtySize()[1]
        self.printHeader(f)

        if self.audio_file.tag and self.handleRemoves(self.audio_file.tag):
            # Reload after removal
            super(ClassicPlugin, self).handleFile(f, tag_version=parse_version)
            if not self.audio_file:
                return

        new_tag = False
        if not self.audio_file.tag:
            self.audio_file.initTag(version=parse_version)
            new_tag = True

        try:
            save_tag = (self.handleEdits(self.audio_file.tag) or
                        self.handlePadding(self.audio_file.tag) or
                        self.args.force_update or self.args.convert_version)
        except ValueError as ex:
            printError(str(ex))
            return

        self.printAudioInfo(self.audio_file.info)

        if not save_tag and new_tag:
            printError(f"No ID3 {id3.versionToString(self.args.tag_version)} tag found!")
            return

        self.printTag(self.audio_file.tag)

        if self.args.write_images_dir:
            for img in self.audio_file.tag.images:
                if img.mime_type not in ImageFrame.URL_MIME_TYPE_VALUES:
                    img_path = "%s%s" % (self.args.write_images_dir,
                                         os.path.sep)
                    if not os.path.isdir(img_path):
                        raise IOError("Directory does not exist: %s" % img_path)
                    img_file = makeUniqueFileName(
                                os.path.join(img_path, img.makeFileName()))
                    printWarning("Writing %s..." % img_file)
                    with open(img_file, "wb") as fp:
                        fp.write(img.image_data)

        if save_tag:
            # Use current tag version unless a convert was supplied
            version = (self.args.convert_version or
                       self.audio_file.tag.version)
            printWarning("Writing ID3 version %s" %
                         id3.versionToString(version))

            # DEFAULT_MAX_PADDING is not set up as argument default,
            # because we don't want to rewrite the file if the user
            # did not trigger that explicitly:
            max_padding = self.args.max_padding
            if max_padding is True:
                max_padding = DEFAULT_MAX_PADDING

            self.audio_file.tag.save(
                    version=version, encoding=self.args.text_encoding,
                    backup=self.args.backup,
                    preserve_file_time=self.args.preserve_file_time,
                    max_padding=max_padding)

        if self.args.rename_pattern:
            # Handle file renaming.
            from eyed3.id3.tag import TagTemplate
            template = TagTemplate(self.args.rename_pattern)
            name = template.substitute(self.audio_file.tag, zeropad=True)
            orig = self.audio_file.path
            try:
                self.audio_file.rename(name)
                printWarning(f"Renamed '{orig}' to '{self.audio_file.path}'")
            except IOError as ex:
                printError(str(ex))

        printMsg(self._getHardRule(self.terminal_width))

    def printHeader(self, file_path):
        printMsg(self._getFileHeader(file_path, self.terminal_width))
        printMsg(self._getHardRule(self.terminal_width))

    def printAudioInfo(self, info):
        if isinstance(info, mp3.Mp3AudioInfo):
            printMsg(boldText("Time: ") +
                     "%s\tMPEG%d, Layer %s\t[ %s @ %s Hz - %s ]" %
                     (formatTime(info.time_secs),
                      info.mp3_header.version,
                      "I" * info.mp3_header.layer,
                      info.bit_rate_str,
                      info.mp3_header.sample_freq, info.mp3_header.mode))
            printMsg(self._getHardRule(self.terminal_width))

    @staticmethod
    def _getDefaultNameForObject(obj_frame, suffix=""):
        if obj_frame.filename:
            name_str = obj_frame.filename
        else:
            name_str = obj_frame.description
            name_str += ".%s" % obj_frame.mime_type.split("/")[1]
        if suffix:
            name_str += suffix
        return name_str

    def printTag(self, tag):
        if isinstance(tag, id3.Tag):
            if self.args.quiet:
                printMsg(f"ID3 {id3.versionToString(tag.version)}: {len(tag.frame_set)} frames")
                return
            printMsg(f"ID3 {id3.versionToString(tag.version)}:")

            artist = tag.artist if tag.artist else ""
            title = tag.title if tag.title else ""
            album = tag.album if tag.album else ""
            printMsg("%s: %s" % (boldText("title"), title))
            printMsg("%s: %s" % (boldText("artist"), artist))
            printMsg("%s: %s" % (boldText("album"), album))
            if tag.album_artist:
                printMsg("%s: %s" % (boldText("album artist"),
                                     tag.album_artist))
            if tag.composer:
                printMsg("%s: %s" % (boldText("composer"), tag.composer))
            if tag.original_artist:
                printMsg("%s: %s" % (boldText("original artist"), tag.original_artist))

            for date, date_label in [
                    (tag.release_date, "release date"),
                    (tag.original_release_date, "original release date"),
                    (tag.recording_date, "recording date"),
                    (tag.encoding_date, "encoding date"),
                    (tag.tagging_date, "tagging date"),
                   ]:
                if date:
                    printMsg("%s: %s" % (boldText(date_label), str(date)))

            track_str = ""
            (track_num, track_total) = tag.track_num
            if track_num is not None:
                track_str = str(track_num)
                if track_total:
                    track_str += "/%d" % track_total

            genre = tag.genre if not self.args.non_std_genres else tag.non_std_genre
            genre_str = f"{boldText('genre')}: {genre.name} (id {genre.id})" if genre else ""
            printMsg(f"{boldText('track')}: {track_str}\t\t{genre_str}")

            (num, total) = tag.disc_num
            if num is not None:
                disc_str = str(num)
                if total:
                    disc_str += "/%d" % total
                printMsg("%s: %s" % (boldText("disc"), disc_str))

            # PCNT
            play_count = tag.play_count
            if tag.play_count is not None:
                printMsg("%s %d" % (boldText("Play Count:"), play_count))

            # POPM
            for popm in tag.popularities:
                printMsg("%s [email: %s] [rating: %d] [play count: %d]" %
                         (boldText("Popularity:"), popm.email, popm.rating,
                          popm.count))

            # TBPM
            bpm = tag.bpm
            if bpm is not None:
                printMsg("%s %d" % (boldText("BPM:"), bpm))

            # TPUB
            pub = tag.publisher
            if pub is not None:
                printMsg("%s %s" % (boldText("Publisher/label:"), pub))

            # UFID
            for ufid in tag.unique_file_ids:
                printMsg("%s [%s] : %s" %
                        (boldText("Unique File ID:"), ufid.owner_id,
                         ufid.uniq_id.decode("unicode_escape")))

            # COMM
            for c in tag.comments:
                printMsg("%s: [Description: %s] [Lang: %s]\n%s" %
                         (boldText("Comment"), c.description or "",
                          c.lang.decode("ascii") or "", c.text or ""))

            # USLT
            for l in tag.lyrics:
                printMsg("%s: [Description: %s] [Lang: %s]\n%s" %
                         (boldText("Lyrics"), l.description or "",
                          l.lang.decode("ascii") or "", l.text))

            # TXXX
            for f in tag.user_text_frames:
                printMsg("%s: [Description: %s]\n%s" %
                         (boldText("UserTextFrame"), f.description, f.text))

            # URL frames
            for desc, url in (("Artist URL", tag.artist_url),
                              ("Audio source URL", tag.audio_source_url),
                              ("Audio file URL", tag.audio_file_url),
                              ("Internet radio URL", tag.internet_radio_url),
                              ("Commercial URL", tag.commercial_url),
                              ("Payment URL", tag.payment_url),
                              ("Publisher URL", tag.publisher_url),
                              ("Copyright URL", tag.copyright_url),
                             ):
                if url:
                    printMsg("%s: %s" % (boldText(desc), url))

            # user url frames
            for u in tag.user_url_frames:
                printMsg("%s [Description: %s]: %s" % (u.id, u.description,
                                                       u.url))

            # APIC
            for img in tag.images:
                if img.mime_type not in ImageFrame.URL_MIME_TYPE_VALUES:
                    printMsg("%s: [Size: %d bytes] [Type: %s]" %
                        (boldText(img.picTypeToString(img.picture_type) +
                                  " Image"),
                        len(img.image_data),
                        img.mime_type))
                    printMsg("Description: %s" % img.description)
                    printMsg("")
                else:
                    printMsg("%s: [Type: %s] [URL: %s]" %
                        (boldText(img.picTypeToString(img.picture_type) +
                                  " Image"),
                        img.mime_type, img.image_url))
                    printMsg("Description: %s" % img.description)
                    printMsg("")

            # GOBJ
            for obj in tag.objects:
                printMsg("%s: [Size: %d bytes] [Type: %s]" %
                         (boldText("GEOB"), len(obj.object_data),
                          obj.mime_type))
                printMsg("Description: %s" % obj.description)
                printMsg("Filename: %s" % obj.filename)
                printMsg("\n")
                if self.args.write_objects_dir:
                    obj_path = "%s%s" % (self.args.write_objects_dir, os.sep)
                    if not os.path.isdir(obj_path):
                        raise IOError("Directory does not exist: %s" % obj_path)
                    obj_file = self._getDefaultNameForObject(obj)
                    count = 1
                    while os.path.exists(os.path.join(obj_path, obj_file)):
                        obj_file = self._getDefaultNameForObject(obj,
                                                                 str(count))
                        count += 1
                    printWarning("Writing %s..." % os.path.join(obj_path,
                                                                obj_file))
                    with open(os.path.join(obj_path, obj_file), "wb") as fp:
                        fp.write(obj.object_data)

            # PRIV
            for p in tag.privates:
                printMsg("%s: [Data: %d bytes]" % (boldText("PRIV"),
                                                   len(p.data)))
                printMsg("Owner Id: %s" % p.owner_id.decode("ascii"))

            # MCDI
            if tag.cd_id:
                printMsg("\n%s: [Data: %d bytes]" % (boldText("MCDI"),
                                                     len(tag.cd_id)))

            # USER
            if tag.terms_of_use:
                printMsg("\nTerms of Use (%s): %s" % (boldText("USER"),
                                                      tag.terms_of_use))

            # --verbose
            if self.args.verbose:
                printMsg(self._getHardRule(self.terminal_width))
                printMsg("%d ID3 Frames:" % len(tag.frame_set))
                for fid in tag.frame_set:
                    frames = tag.frame_set[fid]
                    num_frames = len(frames)
                    count = " x %d" % num_frames if num_frames > 1 else ""
                    if not tag.isV1():
                        total_bytes = sum(
                                tuple(frame.header.data_size + frame.header.size
                                          for frame in frames if frame.header))
                    else:
                        total_bytes = 30
                    if total_bytes:
                        printMsg("%s%s (%d bytes)" % (fid.decode("ascii"),
                                                      count, total_bytes))
                printMsg("%d bytes unused (padding)" %
                         (tag.file_info.tag_padding_size, ))
        else:
            raise TypeError("Unknown tag type: " + str(type(tag)))

    def handleRemoves(self, tag):
        remove_version = 0
        status = False
        rm_str = ""
        if self.args.remove_all:
            remove_version = id3.ID3_ANY_VERSION
            rm_str = "v1.x and/or v2.x"
        elif self.args.remove_v1:
            remove_version = id3.ID3_V1
            rm_str = "v1.x"
        elif self.args.remove_v2:
            remove_version = id3.ID3_V2
            rm_str = "v2.x"

        if remove_version:
            status = id3.Tag.remove(tag.file_info.name, remove_version,
                                    preserve_file_time=self.args.preserve_file_time)
            printWarning(f"Removing ID3 {rm_str} tag: {'SUCCESS' if status else 'FAIL'}")

        return status

    def handlePadding(self, tag):
        max_padding = self.args.max_padding
        if max_padding is None or max_padding is True:
            return False
        padding = tag.file_info.tag_padding_size
        needs_change = padding > max_padding
        return needs_change

    def handleEdits(self, tag):
        retval = False

        # --remove-all-*, Handling removes first means later options are still
        # applied
        for what, arg, fid in (("comments", self.args.remove_all_comments,
                                id3.frames.COMMENT_FID),
                               ("lyrics", self.args.remove_all_lyrics,
                                id3.frames.LYRICS_FID),
                               ("images", self.args.remove_all_images,
                                id3.frames.IMAGE_FID),
                               ("objects", self.args.remove_all_objects,
                                id3.frames.OBJECT_FID),
                               ):
            if arg and tag.frame_set[fid]:
                printWarning("Removing all %s..." % what)
                del tag.frame_set[fid]
                retval = True

        # --artist, --title, etc. All common/simple text frames.
        for (what, setFunc) in (
                ("artist", partial(tag._setArtist, self.args.artist)),
                ("album", partial(tag._setAlbum, self.args.album)),
                ("album artist", partial(tag._setAlbumArtist,
                                         self.args.album_artist)),
                ("title", partial(tag._setTitle, self.args.title)),
                ("genre", partial(tag._setGenre, self.args.genre,
                                  id3_std=not self.args.non_std_genres)),
                ("release date", partial(tag._setReleaseDate,
                                         self.args.release_date)),
                ("original release date", partial(tag._setOrigReleaseDate,
                                                  self.args.orig_release_date)),
                ("recording date", partial(tag._setRecordingDate,
                                           self.args.recording_date)),
                ("encoding date", partial(tag._setEncodingDate,
                                          self.args.encoding_date)),
                ("tagging date", partial(tag._setTaggingDate,
                                         self.args.tagging_date)),
                ("beats per minute", partial(tag._setBpm, self.args.bpm)),
                ("publisher", partial(tag._setPublisher, self.args.publisher)),
                ("composer", partial(tag._setComposer, self.args.composer)),
                ("orig-artist", partial(tag._setOrigArtist, self.args.orig_artist)),
              ):
            if setFunc.args[0] is not None:
                printWarning("Setting %s: %s" % (what, setFunc.args[0]))
                setFunc()
                retval = True

        def _checkNumberedArgTuples(curr, new):
            n = None
            if new not in [(None, None), curr]:
                n = [None] * 2
                for i in (0, 1):
                    if new[i] == 0:
                        n[i] = None
                    else:
                        n[i] = new[i] or curr[i]
                n = tuple(n)
            # Returning None means do nothing, (None, None) would clear both vals
            return n

        # --artist-{city,state,country}
        origin = core.ArtistOrigin(self.args.artist_city,
                                   self.args.artist_state,
                                   self.args.artist_country)
        if origin or (dataclasses.astuple(origin) != (None, None, None) and tag.artist_origin):
            printWarning(f"Setting artist origin: {origin}")
            tag.artist_origin = origin
            retval = True

        # --track, --track-total
        track_info = _checkNumberedArgTuples(tag.track_num,
                                             (self.args.track,
                                              self.args.track_total))
        if track_info is not None:
            printWarning("Setting track info: %s" % str(track_info))
            tag.track_num = track_info
            retval = True

        # --track-offset
        if self.args.track_offset:
            offset = self.args.track_offset
            tag.track_num = (tag.track_num[0] + offset, tag.track_num[1])
            printWarning("%s track info by %d: %d" %
                         ("Incrementing" if offset > 0 else "Decrementing",
                         offset, tag.track_num[0]))
            retval = True

        # --disc-num, --disc-total
        disc_info = _checkNumberedArgTuples(tag.disc_num,
                                            (self.args.disc_num,
                                             self.args.disc_total))
        if disc_info is not None:
            printWarning("Setting disc info: %s" % str(disc_info))
            tag.disc_num = disc_info
            retval = True

        # -Y, --release-year
        if self.args.release_year is not None:
            # empty string means clean, None means not given
            year = self.args.release_year
            printWarning(f"Setting release year: {year}")
            tag.release_date = int(year) if year else None
            retval = True

        # -c , simple comment
        if self.args.simple_comment:
            # Just add it as if it came in --add-comment
            self.args.comments.append((self.args.simple_comment, "",
                                       id3.DEFAULT_LANG))

        # --remove-comment, remove-lyrics, --remove-image, --remove-object
        for what, arg, accessor in (("comment", self.args.remove_comment,
                                     tag.comments),
                                    ("lyrics", self.args.remove_lyrics,
                                     tag.lyrics),
                                    ("image", self.args.remove_image,
                                     tag.images),
                                    ("object", self.args.remove_object,
                                     tag.objects),
                                   ):
            for vals in arg:
                if type(vals) is str:
                    frame = accessor.remove(vals)
                else:
                    frame = accessor.remove(*vals)
                if frame:
                    printWarning("Removed %s %s" % (what, str(vals)))
                    retval = True
                else:
                    printError("Removing %s failed, %s not found" %
                               (what, str(vals)))

        # --add-comment, --add-lyrics
        for what, arg, accessor in (("comment", self.args.comments,
                                     tag.comments),
                                    ("lyrics", self.args.lyrics, tag.lyrics),
                                   ):
            for text, desc, lang in arg:
                printWarning("Setting %s: %s/%s" %
                             (what, desc, str(lang, "ascii")))
                accessor.set(text, desc, b(lang))
                retval = True

        # --play-count
        playcount_arg = self.args.play_count
        if playcount_arg:
            increment, pc = playcount_arg
            if increment:
                printWarning("Increment play count by %d" % pc)
                tag.play_count += pc
            else:
                printWarning("Setting play count to %d" % pc)
                tag.play_count = pc
            retval = True

        # --add-popularity
        for email, rating, play_count in self.args.popularities:
            tag.popularities.set(email.encode("latin1"), rating, play_count)
            retval = True

        # --remove-popularity
        for email in self.args.remove_popularity:
            popm = tag.popularities.remove(email.encode("latin1"))
            if popm:
                retval = True

        # --text-frame, --url-frame
        for what, arg, setter in (
                ("text frame", self.args.text_frames, tag.setTextFrame),
                ("url frame", self.args.url_frames, tag._setUrlFrame),
              ):
            for fid, text in arg:
                if text:
                    printWarning("Setting %s %s to '%s'" % (fid, what, text))
                else:
                    printWarning("Removing %s %s" % (fid, what))
                setter(fid, text)
                retval = True

        # --user-text-frame, --user-url-frame
        for what, arg, accessor in (
                ("user text frame", self.args.user_text_frames,
                 tag.user_text_frames),
                ("user url frame", self.args.user_url_frames,
                 tag.user_url_frames),
              ):
            for desc, text in arg:
                if text:
                    printWarning(f"Setting '{desc}' {what} to '{text}'")
                    accessor.set(text, desc)
                else:
                    printWarning(f"Removing '{desc}' {what}")
                    accessor.remove(desc)
                retval = True

        # --add-image
        for img_path, img_type, img_mt, img_desc in self.args.images:
            assert img_path
            printWarning("Adding image %s" % img_path)
            if img_mt not in ImageFrame.URL_MIME_TYPE_VALUES:
                with open(img_path, "rb") as img_fp:
                    tag.images.set(img_type, img_fp.read(), img_mt, img_desc)
            else:
                tag.images.set(img_type, None, None, img_desc, img_url=img_path)
            retval = True

        # --add-object
        for obj_path, obj_mt, obj_desc, obj_fname in self.args.objects or []:
            assert obj_path
            printWarning("Adding object %s" % obj_path)
            with open(obj_path, "rb") as obj_fp:
                tag.objects.set(obj_fp.read(), obj_mt, obj_desc, obj_fname)
            retval = True

        # --unique-file-id
        for arg in self.args.unique_file_ids:
            owner_id, id = arg
            if not id:
                if tag.unique_file_ids.remove(owner_id):
                    printWarning("Removed unique file ID '%s'" % owner_id)
                    retval = True
                else:
                    printWarning("Unique file ID '%s' not found" % owner_id)
            else:
                tag.unique_file_ids.set(id, owner_id.encode("latin1"))
                printWarning("Setting unique file ID '%s' to %s" %
                              (owner_id, id))
                retval = True

        # --remove-frame
        for fid in self.args.remove_fids:
            assert(isinstance(fid, bytes))
            if fid in tag.frame_set:
                del tag.frame_set[fid]
                retval = True

        return retval


def _getTemplateKeys():
    keys = list(id3.TagTemplate("")._makeMapping(None, False).keys())
    keys.sort()
    return ", ".join(["$%s" % v for v in keys])


ARGS_HELP = {
        "--artist": "Set the artist name.",
        "--album": "Set the album name.",
        "--album-artist": "Set the album artist name. '%s', for example. "
                          "Another example is collaborations when the "
                          "track artist might be 'Eminem featuring Proof' "
                          "the album artist would be 'Eminem'." %
                          core.VARIOUS_ARTISTS,
        "--title": "Set the track title.",
        "--track": "Set the track number. Use 0 to clear.",
        "--track-total": "Set total number of tracks. Use 0 to clear.",
        "--disc-num": "Set the disc number. Use 0 to clear.",
        "--disc-total": "Set total number of discs in set. Use 0 to clear.",
        "--genre": "Set the genre. If the argument is a standard ID3 genre "
                   "name or number both will be set. Otherwise, any string "
                   "can be used. Run 'eyeD3 --plugin=genres' for a list of "
                   "standard ID3 genre names/ids.",
        "--non-std-genres": "Disables certain ID3 genre standards, such as the "
                            "mapping of numeric value to genre names. For example, "
                            "genre=1 is taken literally, not mapped to 'Classic Rock'.",
        "--release-year": "Set the year the track was released. Use the date "
                          "options for more precise values or dates other "
                          "than release.",

        "--v1": "Only read and write ID3 v1.x tags. By default, v1.x tags are "
                "only read or written if there is not a v2 tag in the file.",
        "--v2": "Only read/write ID3 v2.x tags. This is the default unless "
                "the file only contains a v1 tag.",

        "--to-v1.1": "Convert the file's tag to ID3 v1.1 (Or 1.0 if there is "
                     "no track number)",
        "--to-v2.3": "Convert the file's tag to ID3 v2.3",
        "--to-v2.4": "Convert the file's tag to ID3 v2.4",

        "--release-date": "Set the date the track/album was released",
        "--orig-release-date": "Set the original date the track/album was "
                               "released",
        "--recording-date": "Set the date the track/album was recorded",
        "--encoding-date": "Set the date the file was encoded",
        "--tagging-date": "Set the date the file was tagged",

        "--comment": "Set a comment. In ID3 tags this is the comment with "
                     "an empty description. See --add-comment to add multiple "
                     "comment frames.",
        "--add-comment":
          "Add or replace a comment. There may be more than one comment in a "
          "tag, as long as the DESCRIPTION and LANG values are unique. The "
          "default DESCRIPTION is '' and the default language code is '%s'." %
          str(id3.DEFAULT_LANG, "ascii"),
        "--remove-comment": "Remove comment matching DESCRIPTION and LANG. "
                            "The default language code is '%s'." %
                            str(id3.DEFAULT_LANG, "ascii"),
        "--remove-all-comments": "Remove all comments from the tag.",

        "--add-lyrics":
          "Add or replace a lyrics. There may be more than one set of lyrics "
          "in a tag, as long as the DESCRIPTION and LANG values are unique. "
          "The default DESCRIPTION is '' and the default language code is "
          "'%s'." % str(id3.DEFAULT_LANG, "ascii"),
        "--remove-lyrics": "Remove lyrics matching DESCRIPTION and LANG. "
                            "The default language code is '%s'." %
                            str(id3.DEFAULT_LANG, "ascii"),
        "--remove-all-lyrics": "Remove all lyrics from the tag.",

        "--publisher": "Set the publisher/label name",
        "--play-count": "Set the number of times played counter. If the "
                        "argument value begins with '+' the tag's play count "
                        "is incremented by N, otherwise the value is set to "
                        "exactly N.",
        "--bpm": "Set the beats per minute value.",

        "--text-frame": "Set the value of a text frame. To remove the "
                        "frame, specify an empty value. For example, "
                        "--text-frame='TDRC:'",
        "--user-text-frame": "Set the value of a user text frame (i.e., TXXX). "
                             "To remove the frame, specify an empty value. "
                             "e.g., --user-text-frame='SomeDesc:'",
        "--url-frame": "Set the value of a URL frame. To remove the frame, "
                       "specify an empty value. e.g., --url-frame='WCOM:'",
        "--user-url-frame": "Set the value of a user URL frame (i.e., WXXX). "
                            "To remove the frame, specify an empty value. "
                            "e.g., --user-url-frame='SomeDesc:'",

        "--add-image": "Add or replace an image. There may be more than one "
                       "image in a tag, as long as the DESCRIPTION values are "
                       "unique. The default DESCRIPTION is ''. If PATH begins "
                       "with 'http[s]://' then it is interpreted as a URL "
                       "instead of a file containing image data. The TYPE must "
                       "be one of the following: %s."
                       % (", ".join([ImageFrame.picTypeToString(t)
                                    for t in range(ImageFrame.MIN_TYPE,
                                                   ImageFrame.MAX_TYPE + 1)]),
                         ),
        "--remove-image": "Remove image matching DESCRIPTION.",
        "--remove-all-images": "Remove all images from the tag",
        "--write-images": "Causes all attached images (APIC frames) to be "
                          "written to the specified directory.",

        "--add-object": "Add or replace an object. There may be more than one "
                        "object in a tag, as long as the DESCRIPTION values "
                        "are unique. The default DESCRIPTION is ''.",
        "--remove-object": "Remove object matching DESCRIPTION.",
        "--remove-all-objects": "Remove all objects from the tag",
        "--write-objects": "Causes all attached objects (GEOB frames) to be "
                           "written to the specified directory.",

        "--add-popularity": "Adds a pupularity metric. There may be multiples "
                           "popularity values, but each must have a unique "
                           "email address component. The rating is a number "
                           "between 0 (worst) and 255 (best). The play count "
                           "is optional, and defaults to 0, since there is "
                           "already a dedicated play count frame.",
        "--remove-popularity": "Removes the popularity frame with the "
                               "specified email key.",

        "--remove-v1": "Remove ID3 v1.x tag.",
        "--remove-v2": "Remove ID3 v2.x tag.",
        "--remove-all": "Remove ID3 v1.x and v2.x tags.",

        "--remove-frame": "Remove all frames with the given ID. This option "
                          "may be specified multiple times.",

        "--max-padding": "Shrink file if tag padding (unused space) exceeds "
                         "the given number of bytes. "
                         "(Useful e.g. after removal of large cover art.) "
                         "Default is 64 KiB, file will be rewritten with "
                         "default padding (1 KiB) or max padding, whichever "
                         "is smaller.",
        "--no-max-padding": "Disable --max-padding altogether.",

        "--force-update": "Rewrite the tag despite there being no edit "
                          "options.",
        "--verbose": "Show all available tag data",
        "--unique-file-id": "Add a unique file ID frame. If the ID arg is "
                            "empty the frame is removed. An OWNER_ID is "
                            "required. The ID may be no more than 64 bytes.",
        "--encoding": "Set the encoding that is used for all text frames. "
                      "This option is only applied if the tag is updated "
                      "as the result of an edit option (e.g. --artist, "
                       "--title, etc.) or --force-update is specified.",
        "--rename": "Rename file (the extension is not affected) "
                    "based on data in the tag using substitution "
                    "variables: " + _getTemplateKeys(),
        "--preserve-file-times": "When writing, do not update file "
                                 "modification times.",
        "--track-offset": "Increment/decrement the track number by [-]N. "
                          "This option is applied after --track=N is set.",
        "--composer": "Set the composer's name.",
        "--orig-artist": "Set the orignal artist's name. For example, a cover song can include "
                         "the orignal author of the track.",
}



from pylast import SIZE_EXTRA_LARGE, SIZE_LARGE, SIZE_MEDIUM, SIZE_MEGA, SIZE_SMALL
from pylast import LastFMNetwork, WSError

api_k = "a5f0ac61e7db2481b054ba52ff9a654f"
api_s = "0c4a52ae5dcdbba1f9e782833a50b623"
_network = None


def Client():
    global _network
    if not _network:
        _network = LastFMNetwork(api_key=api_k, api_secret=api_s)
        _network.enable_rate_limit()
    return _network


def getArtist(artist):
    return Client().get_artist(artist)


def getAlbum(artist, title):
    return Client().get_album(artist, title)


def getAlbumArt(artist, title, size=SIZE_EXTRA_LARGE):
    return _getArt(getAlbum(artist, title), size=size)


def getArtistArt(artist, size=SIZE_EXTRA_LARGE):
    return _getArt(getArtist(artist), size=size)


def _getArt(obj, size=SIZE_EXTRA_LARGE):
    try:
        return obj.get_cover_image(size)
    except WSError:
        raise ValueError("{} not found.".format(obj.__class__.__name__))


if __name__ == "__main__":
    album = getAlbum("Melvins", "Houdini")
    for sz in (SIZE_SMALL, SIZE_MEGA, SIZE_MEDIUM, SIZE_LARGE,
               SIZE_EXTRA_LARGE):
        print(album.get_cover_image(sz))

    melvins = getArtist("Melvins")
    print(melvins)
    for sz in (SIZE_SMALL, SIZE_MEGA, SIZE_MEDIUM, SIZE_LARGE,
               SIZE_EXTRA_LARGE):
        print(melvins.get_cover_image(sz))



import os
import string
import shutil
import tempfile
import textwrap
from codecs import ascii_encode


from ..utils import requireUnicode, chunkCopy, datePicker, b
from .. import core
from ..core import TXXX_ALBUM_TYPE, TXXX_ARTIST_ORIGIN, ALBUM_TYPE_IDS, ArtistOrigin
from .. import Error
from . import (ID3_ANY_VERSION, ID3_DEFAULT_VERSION, ID3_V1, ID3_V1_0, ID3_V1_1,
               ID3_V2, ID3_V2_2, ID3_V2_3, ID3_V2_4, versionToString)
from . import DEFAULT_LANG
from . import Genre
from . import frames
from .headers import TagHeader, ExtendedTagHeader

from ..utils.log import getLogger
log = getLogger(__name__)

ID3_V1_COMMENT_DESC = "ID3v1.x Comment"
ID3_V1_MAX_TEXTLEN = 30
ID3_V1_STRIP_CHARS = string.whitespace.encode("latin1") + b"\x00"
DEFAULT_PADDING = 256


class TagException(Error):
    pass


class Tag(core.Tag):
    def __init__(self, version=ID3_DEFAULT_VERSION, **kwargs):
        self.file_info = None
        self.header = None
        self.extended_header = None
        self.frame_set = None

        self._comments = None
        self._images = None
        self._lyrics = None
        self._objects = None
        self._privates = None
        self._user_texts = None
        self._unique_file_ids = None
        self._user_urls = None
        self._chapters = None
        self._tocs = None
        self._popularities = None

        self.file_info = None
        self.clear(version=version)
        super().__init__(**kwargs)

    def clear(self, *, version=ID3_DEFAULT_VERSION):
        """Reset all tag data."""
        # ID3 tag header
        self.header = TagHeader(version=version)
        # Optional extended header in v2 tags.
        self.extended_header = ExtendedTagHeader()
        # Contains the tag's frames. ID3v1 fields are read and converted
        #  the the corresponding v2 frame.
        self.frame_set = frames.FrameSet()
        self._comments = CommentsAccessor(self.frame_set)
        self._images = ImagesAccessor(self.frame_set)
        self._lyrics = LyricsAccessor(self.frame_set)
        self._objects = ObjectsAccessor(self.frame_set)
        self._privates = PrivatesAccessor(self.frame_set)
        self._user_texts = UserTextsAccessor(self.frame_set)
        self._unique_file_ids = UniqueFileIdAccessor(self.frame_set)
        self._user_urls = UserUrlsAccessor(self.frame_set)
        self._chapters = ChaptersAccessor(self.frame_set)
        self._tocs = TocAccessor(self.frame_set)
        self._popularities = PopularitiesAccessor(self.frame_set)

    def parse(self, fileobj, version=ID3_ANY_VERSION):
        self.clear()
        version = version or ID3_ANY_VERSION

        close_file = False
        try:
            filename = fileobj.name
        except AttributeError:
            if type(fileobj) is str:
                filename = fileobj
                fileobj = open(filename, "rb")
                close_file = True
            else:
                raise ValueError(f"Invalid type: {type(fileobj)}")

        self.file_info = FileInfo(filename)

        try:
            tag_found = False
            padding = 0
            # The & is for supporting the "meta" versions, any, etc.
            if version[0] & 2:
                tag_found, padding = self._loadV2Tag(fileobj)

            if not tag_found and version[0] & 1:
                tag_found, padding = self._loadV1Tag(fileobj)
                if tag_found:
                    self.extended_header = None

            if tag_found and self.isV2:
                self.file_info.tag_size = (TagHeader.SIZE +
                                           self.header.tag_size)
            if tag_found:
                self.file_info.tag_padding_size = padding

        finally:
            if close_file:
                fileobj.close()

        return tag_found

    def _loadV2Tag(self, fp):
        """Returns (tag_found, padding_len)"""
        fp.seek(0)

        # Look for a tag and if found load it.
        if not self.header.parse(fp):
            return False, 0

        # Read the extended header if present.
        if self.header.extended:
            self.extended_header.parse(fp, self.header.version)

        # Header is definitely there so at least one frame *must* follow.
        padding = self.frame_set.parse(fp, self.header,
                                       self.extended_header)

        log.debug("Tag contains %d bytes of padding." % padding)
        return True, padding

    def _loadV1Tag(self, fp):
        v1_enc = "latin1"

        # Seek to the end of the file where all v1x tags are written.
        # v1.x tags are 128 bytes min and max
        fp.seek(0, 2)
        if fp.tell() < 128:
            return False, 0
        fp.seek(-128, 2)
        tag_data = fp.read(128)

        if tag_data[0:3] != b"TAG":
            return False, 0

        log.debug("Located ID3 v1 tag")
        # v1.0 is implied until a v1.1 feature is recognized.
        self.version = ID3_V1_0

        title = tag_data[3:33].strip(ID3_V1_STRIP_CHARS)
        log.debug("Title: %s" % title)
        if title:
            self.title = str(title, v1_enc)

        artist = tag_data[33:63].strip(ID3_V1_STRIP_CHARS)
        log.debug("Artist: %s" % artist)
        if artist:
            self.artist = str(artist, v1_enc)

        album = tag_data[63:93].strip(ID3_V1_STRIP_CHARS)
        log.debug("Album: %s" % album)
        if album:
            self.album = str(album, v1_enc)

        year = tag_data[93:97].strip(ID3_V1_STRIP_CHARS)
        log.debug("Year: %s" % year)
        try:
            if year and int(year):
                # Values here typically mean the year of release
                self.release_date = int(year)
        except ValueError:
            # Bogus year strings.
            log.warn("ID3v1.x tag contains invalid year: %s" % year)
            pass

        # Can't use ID3_V1_STRIP_CHARS here, since the final byte is numeric
        comment = tag_data[97:127].rstrip(b"\x00")
        # Track numbers stuffed in the comment field is what makes v1.1
        if comment:
            if (len(comment) >= 2 and
                    # Python the slices (the chars), so this is really
                    # comment[2]       and        comment[-1]
                    comment[-2:-1] == b"\x00"):
                log.debug("Track Num found, setting version to v1.1")
                self.version = ID3_V1_1

                track = comment[-1]
                self.track_num = (track, None)
                log.debug("Track: " + str(track))
                comment = comment[:-2].strip(ID3_V1_STRIP_CHARS)

            # There may only have been a track #
            if comment:
                log.debug(f"Comment: {comment}")
                self.comments.set(str(comment, v1_enc), ID3_V1_COMMENT_DESC)

        genre = ord(tag_data[127:128])
        log.debug(f"Genre ID: {genre}")
        try:
            self.genre = genre
        except ValueError as ex:
            log.warning(ex)
            self.genre = None

        return True, 0

    @property
    def version(self):
        return self.header.version

    @version.setter
    def version(self, v):
        # Tag version changes required possible frame conversion
        std, non = self._checkForConversions(v)
        converted = []
        if non:
            converted = self._convertFrames(std, non, v)
        if converted:
            self.frame_set.clear()
            for frame in (std + converted):
                self.frame_set[frame.id] = frame

        self.header.version = v

    def isV1(self):
        """Test ID3 major version for v1.x"""
        return self.header.major_version == 1

    def isV2(self):
        """Test ID3 major version for v2.x"""
        return self.header.major_version == 2

    @requireUnicode(2)
    def setTextFrame(self, fid: bytes, txt: str):
        fid = b(fid, ascii_encode)
        if not fid.startswith(b"T") or fid.startswith(b"TX"):
            raise ValueError("Invalid frame-id for text frame")

        if not txt and self.frame_set[fid]:
            del self.frame_set[fid]
        elif txt:
            self.frame_set.setTextFrame(fid, txt)

    # FIXME: is returning data not a Frame.
    def getTextFrame(self, fid: bytes):
        fid = b(fid, ascii_encode)
        if not fid.startswith(b"T") or fid.startswith(b"TX"):
            raise ValueError("Invalid frame-id for text frame")
        f = self.frame_set[fid]
        return f[0].text if f else None

    @requireUnicode(1)
    def _setArtist(self, val):
        self.setTextFrame(frames.ARTIST_FID, val)

    def _getArtist(self):
        return self.getTextFrame(frames.ARTIST_FID)

    @requireUnicode(1)
    def _setAlbumArtist(self, val):
        self.setTextFrame(frames.ALBUM_ARTIST_FID, val)

    def _getAlbumArtist(self):
        return self.getTextFrame(frames.ALBUM_ARTIST_FID)

    @requireUnicode(1)
    def _setComposer(self, val):
        self.setTextFrame(frames.COMPOSER_FID, val)

    def _getComposer(self):
        return self.getTextFrame(frames.COMPOSER_FID)

    @property
    def composer(self):
        return self._getComposer()

    @composer.setter
    def composer(self, v):
        self._setComposer(v)

    @requireUnicode(1)
    def _setAlbum(self, val):
        self.setTextFrame(frames.ALBUM_FID, val)

    def _getAlbum(self):
        return self.getTextFrame(frames.ALBUM_FID)

    @requireUnicode(1)
    def _setTitle(self, val):
        self.setTextFrame(frames.TITLE_FID, val)

    def _getTitle(self):
        return self.getTextFrame(frames.TITLE_FID)

    def _setTrackNum(self, val):
        self._setNum(frames.TRACKNUM_FID, val)

    def _getTrackNum(self):
        return self._splitNum(frames.TRACKNUM_FID)

    def _setDiscNum(self, val):
        self._setNum(frames.DISCNUM_FID, val)

    def _getDiscNum(self):
        return self._splitNum(frames.DISCNUM_FID)

    def _splitNum(self, fid):
        f = self.frame_set[fid]
        first, second = None, None
        if f and f[0].text:
            n = f[0].text.split('/')
            try:
                first = int(n[0])
                second = int(n[1]) if len(n) == 2 else None
            except ValueError as ex:
                log.warning(str(ex))
        return first, second

    def _setNum(self, fid, val):
        if type(val) is str:
            val = int(val)

        if type(val) is tuple:
            if len(val) != 2:
                raise ValueError("A 2-tuple of int values is required.")
            else:
                tn, tt = tuple([int(v) if v is not None else None for v in val])
        elif type(val) is int:
            tn, tt = val, None
        elif val is None:
            tn, tt = None, None
        else:
            raise TypeError("Invalid value, should int 2-tuple, int, or None: "
                            f"{val} ({val.__class__.__name__})")

        n = (tn, tt)

        if n[0] is None and n[1] is None:
            if self.frame_set[fid]:
                del self.frame_set[fid]
            return

        total_str = ""
        if n[1] is not None:
            if 0 <= n[1] <= 9:
                total_str = "0" + str(n[1])
            else:
                total_str = str(n[1])

        t = n[0] if n[0] else 0
        track_str = str(t)

        # Pad with zeros according to how large the total count is.
        if len(track_str) == 1:
            track_str = "0" + track_str
        if len(track_str) < len(total_str):
            track_str = ("0" * (len(total_str) - len(track_str))) + track_str

        final_str = ""
        if track_str and total_str:
            final_str = "%s/%s" % (track_str, total_str)
        elif track_str and not total_str:
            final_str = track_str

        self.frame_set.setTextFrame(fid, str(final_str))

    @property
    def comments(self):
        return self._comments

    def _getBpm(self):
        from decimal import Decimal, ROUND_HALF_UP, InvalidOperation

        bpm = None
        if frames.BPM_FID in self.frame_set:
            bpm_str = self.frame_set[frames.BPM_FID][0].text or "0"
            try:
                # Round floats since the spec says this is an integer. Python3
                # changed how 'round' works, hence the using of decimal
                bpm = int(Decimal(bpm_str).quantize(1, ROUND_HALF_UP))
            except (InvalidOperation, ValueError) as ex:
                log.warning(ex)
        return bpm

    def _setBpm(self, bpm):
        assert(bpm >= 0)
        self.setTextFrame(frames.BPM_FID, str(bpm))

    bpm = property(_getBpm, _setBpm)

    @property
    def play_count(self):
        if frames.PLAYCOUNT_FID in self.frame_set:
            pc = self.frame_set[frames.PLAYCOUNT_FID][0]
            return pc.count
        else:
            return None

    @play_count.setter
    def play_count(self, count):
        if count is None:
            del self.frame_set[frames.PLAYCOUNT_FID]
            return

        if count < 0:
            raise ValueError("Invalid play count value: %d" % count)

        if self.frame_set[frames.PLAYCOUNT_FID]:
            pc = self.frame_set[frames.PLAYCOUNT_FID][0]
            pc.count = count
        else:
            self.frame_set[frames.PLAYCOUNT_FID] = \
                frames.PlayCountFrame(count=count)

    def _getPublisher(self):
        if frames.PUBLISHER_FID in self.frame_set:
            pub = self.frame_set[frames.PUBLISHER_FID]
            return pub[0].text
        else:
            return None

    @requireUnicode(1)
    def _setPublisher(self, p):
        self.setTextFrame(frames.PUBLISHER_FID, p)

    publisher = property(_getPublisher, _setPublisher)

    @property
    def cd_id(self):
        if frames.CDID_FID in self.frame_set:
            return self.frame_set[frames.CDID_FID][0].toc
        else:
            return None

    @cd_id.setter
    def cd_id(self, toc):
        if len(toc) > 804:
            raise ValueError("CD identifier table of contents can be no "
                             "greater than 804 bytes")

        if self.frame_set[frames.CDID_FID]:
            cdid = self.frame_set[frames.CDID_FID][0]
            cdid.toc = bytes(toc)
        else:
            self.frame_set[frames.CDID_FID] = \
                frames.MusicCDIdFrame(toc=toc)

    @property
    def images(self):
        return self._images

    def _getEncodingDate(self):
        return self._getDate(b"TDEN")

    def _setEncodingDate(self, date):
        self._setDate(b"TDEN", date)
    encoding_date = property(_getEncodingDate, _setEncodingDate)

    @property
    def best_release_date(self):
        """This method tries its best to return a date of some sort, amongst
        alll the possible date frames. The order of preference for a release
        date is 1) date of original release 2) date of this versions release
        3) the recording date. Or None is returned."""
        import warnings
        warnings.warn("Use Tag.getBestDate() instead", DeprecationWarning,
                      stacklevel=2)
        return (self.original_release_date or
                self.release_date or
                self.recording_date)

    def getBestDate(self, prefer_recording_date=False):
        """This method returns a date of some sort, amongst all the possible
        date frames. The order of preference is:

        1) date of original release
        2) date of this versions release
        3) the recording date.

        Unless ``prefer_recording_date`` is ``True`` in which case the order is
        3, 1, 2.

        ``None`` will be returned if no dates are available."""
        return datePicker(self, prefer_recording_date)

    def _getReleaseDate(self):
        if self.version == ID3_V2_3:
            # v2.3 does NOT have a release date, only TORY, so that is what is returned
            return self._getV23OriginalReleaseDate()
        else:
            return self._getDate(b"TDRL")

    def _setReleaseDate(self, date):
        if self.version == ID3_V2_3:
            # v2.3 does NOT have a release date, only TORY, so that is what is set
            self._setOriginalReleaseDate(date)
        else:
            self._setDate(b"TDRL", date)

    release_date = property(_getReleaseDate, _setReleaseDate)
    release_date.__doc__ = textwrap.dedent("""
    The date the audio was released. This is NOT the original date the
    work was released, instead it is more like the pressing or version of the
    release. Original release date is usually what is intended but many programs
    use this frame and/or don't distinguish between the two.

    NOTE: ID3v2.3 only has original release date, so setting release_date is the same as
    original_release_value; they both set TORY.
    """)

    def _getOrigReleaseDate(self):
        if self.version == ID3_V2_3:
            return self._getV23OriginalReleaseDate()
        else:
            return self._getDate(b"TDOR") or self._getV23OriginalReleaseDate()
    _getOriginalReleaseDate = _getOrigReleaseDate

    def _setOrigReleaseDate(self, date):
        if self.version == ID3_V2_3:
            self._setDate(b"TORY", date)
        else:
            self._setDate(b"TDOR", date)
    _setOriginalReleaseDate = _setOrigReleaseDate

    original_release_date = property(_getOrigReleaseDate, _setOrigReleaseDate)
    original_release_date.__doc__ = textwrap.dedent("""
    The date the work was originally released.

    NOTE: ID3v2.3 only stores year. If the Date object is more precise it is store in `XDOR`, and
    XDOR is preferred when acessing. The year-only date is stored in the standard `TORY` frame as
    well.
    """)

    def _getRecordingDate(self):
        if self.version == ID3_V2_3:
            return self._getV23RecordingDate()
        else:
            return self._getDate(b"TDRC")

    def _setRecordingDate(self, date):
        if date in (None, ""):
            for fid in (b"TDRC", b"TYER", b"TDAT", b"TIME"):
                self._setDate(fid, None)
        elif self.version == ID3_V2_4:
            self._setDate(b"TDRC", date)
        else:
            if not isinstance(date, core.Date):
                date = core.Date.parse(date)
            self._setDate(b"TYER", str(date.year))
            if None not in (date.month, date.day):
                date_str = "%s%s" % (str(date.day).rjust(2, "0"),
                                     str(date.month).rjust(2, "0"))
                self._setDate(b"TDAT", date_str)
            if None not in (date.hour, date.minute):
                date_str = "%s%s" % (str(date.hour).rjust(2, "0"),
                                     str(date.minute).rjust(2, "0"))
                self._setDate(b"TIME", date_str)

    recording_date = property(_getRecordingDate, _setRecordingDate)
    """The date of the recording. Many applications use this for release date
    regardless of the fact that this value is rarely known, and release dates
    are more correct."""

    def _getV23RecordingDate(self):
        # v2.3 TYER (yyyy), TDAT (DDMM), TIME (HHmm)
        date = None
        try:
            date_str = b""
            if b"TYER" in self.frame_set:
                date_str = self.frame_set[b"TYER"][0].text.encode("latin1")
                date = core.Date.parse(date_str)
            if b"TDAT" in self.frame_set:
                text = self.frame_set[b"TDAT"][0].text.encode("latin1")
                date_str += b"-%s-%s" % (text[2:], text[:2])
                date = core.Date.parse(date_str)
            if b"TIME" in self.frame_set:
                text = self.frame_set[b"TIME"][0].text.encode("latin1")
                date_str += b"T%s:%s" % (text[:2], text[2:])
                date = core.Date.parse(date_str)
        except ValueError as ex:
            log.warning("Invalid v2.3 TYER, TDAT, or TIME frame: %s" % ex)

        return date

    def _getV23OriginalReleaseDate(self):
        date, date_str = None, None
        try:
            # XDOR is preferred since it can gave a full date, whereas TORY is year only.
            for fid in (b"XDOR", b"TORY"):
                if fid in self.frame_set:
                    date_str = self.frame_set[fid][0].text.encode("latin1")
                    break
            if date_str:
                date = core.Date.parse(date_str)
        except ValueError as ex:
            log.warning(f"Invalid v2.3 TORY/XDOR frame: {ex}")

        return date

    def _getTaggingDate(self):
        return self._getDate(b"TDTG")

    def _setTaggingDate(self, date):
        self._setDate(b"TDTG", date)
    tagging_date = property(_getTaggingDate, _setTaggingDate)

    def _setDate(self, fid, date):
        def removeFrame(frame_id):
            try:
                del self.frame_set[frame_id]
            except KeyError:
                pass

        def setFrame(frame_id, date_val):
            if frame_id in self.frame_set:
                self.frame_set[frame_id][0].date = date_val
            else:
                self.frame_set[frame_id] = frames.DateFrame(frame_id, str(date_val))

        assert fid in frames.DATE_FIDS or fid in frames.DEPRECATED_DATE_FIDS
        if fid == b"XDOR":
            raise ValueError("Set TORY with a full date (i.e. more than year)")

        clean_fids = [fid]
        if fid == b"TORY":
            clean_fids.append(b"XDOR")

        if date in (None, ""):
            for cid in clean_fids:
                removeFrame(cid)
            return

        # Special casing the conversion to DATE objects cuz TDAT and TIME won't
        if fid not in (b"TDAT", b"TIME"):
            # Convert to ISO format which is what FrameSet wants.
            date_type = type(date)
            if date_type is int:
                # The integer year
                date = core.Date(date)
            elif date_type is str:
                date = core.Date.parse(date)
            elif not isinstance(date, core.Date):
                raise TypeError(f"Invalid type: {date_type}")

        if fid == b"TORY":
            setFrame(fid, date.year)
            if date.month:
                setFrame(b"XDOR", date)
            else:
                removeFrame(b"XDOR")
        else:
            setFrame(fid, date)

    def _getDate(self, fid):
        if fid in (b"TORY", b"XDOR"):
            return self._getV23OriginalReleaseDate()

        if fid in self.frame_set:
            if fid in (b"TYER", b"TDAT", b"TIME"):
                if fid == b"TYER":
                    # Contain years only, date conversion can happen
                    return core.Date(int(self.frame_set[fid][0].text))
                else:
                    return self.frame_set[fid][0].text
            else:
                return self.frame_set[fid][0].date
        else:
            return None

    @property
    def lyrics(self):
        return self._lyrics

    @property
    def disc_num(self):
        return self._getDiscNum()

    @disc_num.setter
    def disc_num(self, val):
        self._setDiscNum(val)

    @property
    def objects(self):
        return self._objects

    @property
    def privates(self):
        return self._privates

    @property
    def popularities(self):
        return self._popularities

    def _getGenre(self, id3_std=True):
        f = self.frame_set[frames.GENRE_FID]
        if f and f[0].text:
            try:
                return Genre.parse(f[0].text, id3_std=id3_std)
            except ValueError:  # pragma: nocover
                return None
        else:
            return None

    def _setGenre(self, g, id3_std=True):
        """Set the genre.
        Four types are accepted for the ``g`` argument.
        A Genre object, an acceptable (see Genre.parse) genre string,
        or an integer genre ID all will set the value. A value of None will
        remove the genre."""
        if g in ("", None):
            if self.frame_set[frames.GENRE_FID]:
                del self.frame_set[frames.GENRE_FID]
            return

        if isinstance(g, str):
            g = Genre.parse(g, id3_std=id3_std)
        elif isinstance(g, int):
            g = Genre(id=g)
        elif not isinstance(g, Genre):
            raise TypeError(f"Invalid genre data type: {type(g)}")

        assert g
        self.frame_set.setTextFrame(frames.GENRE_FID, f"{g.name if g.name else g.id}")

    # genre property
    genre = property(_getGenre, _setGenre)

    def _getNonStdGenre(self):
        return self._getGenre(id3_std=False)

    def _setNonStdGenre(self, val):
        self._setGenre(val, id3_std=False)

    # non-standard genre (unparsed, unmapped) property
    non_std_genre = property(_getNonStdGenre, _setNonStdGenre)

    @property
    def user_text_frames(self):
        return self._user_texts

    def _setUrlFrame(self, fid, url):
        if fid not in frames.URL_FIDS:
            raise ValueError("Invalid URL frame-id")

        if self.frame_set[fid]:
            if not url:
                del self.frame_set[fid]
            else:
                self.frame_set[fid][0].url = url
        else:
            self.frame_set[fid] = frames.UrlFrame(fid, url)

    def _getUrlFrame(self, fid):
        if fid not in frames.URL_FIDS:
            raise ValueError("Invalid URL frame-id")
        f = self.frame_set[fid]
        return f[0].url if f else None

    @property
    def commercial_url(self):
        return self._getUrlFrame(frames.URL_COMMERCIAL_FID)

    @commercial_url.setter
    def commercial_url(self, url):
        self._setUrlFrame(frames.URL_COMMERCIAL_FID, url)

    @property
    def copyright_url(self):
        return self._getUrlFrame(frames.URL_COPYRIGHT_FID)

    @copyright_url.setter
    def copyright_url(self, url):
        self._setUrlFrame(frames.URL_COPYRIGHT_FID, url)

    @property
    def audio_file_url(self):
        return self._getUrlFrame(frames.URL_AUDIOFILE_FID)

    @audio_file_url.setter
    def audio_file_url(self, url):
        self._setUrlFrame(frames.URL_AUDIOFILE_FID, url)

    @property
    def audio_source_url(self):
        return self._getUrlFrame(frames.URL_AUDIOSRC_FID)

    @audio_source_url.setter
    def audio_source_url(self, url):
        self._setUrlFrame(frames.URL_AUDIOSRC_FID, url)

    @property
    def artist_url(self):
        return self._getUrlFrame(frames.URL_ARTIST_FID)

    @artist_url.setter
    def artist_url(self, url):
        self._setUrlFrame(frames.URL_ARTIST_FID, url)

    @property
    def internet_radio_url(self):
        return self._getUrlFrame(frames.URL_INET_RADIO_FID)

    @internet_radio_url.setter
    def internet_radio_url(self, url):
        self._setUrlFrame(frames.URL_INET_RADIO_FID, url)

    @property
    def payment_url(self):
        return self._getUrlFrame(frames.URL_PAYMENT_FID)

    @payment_url.setter
    def payment_url(self, url):
        self._setUrlFrame(frames.URL_PAYMENT_FID, url)

    @property
    def publisher_url(self):
        return self._getUrlFrame(frames.URL_PUBLISHER_FID)

    @publisher_url.setter
    def publisher_url(self, url):
        self._setUrlFrame(frames.URL_PUBLISHER_FID, url)

    @property
    def user_url_frames(self):
        return self._user_urls

    @property
    def unique_file_ids(self):
        return self._unique_file_ids

    @property
    def terms_of_use(self):
        if self.frame_set[frames.TOS_FID]:
            return self.frame_set[frames.TOS_FID][0].text

    @terms_of_use.setter
    def terms_of_use(self, tos):
        """Set the terms of use text.
        To specify a language (other than DEFAULT_LANG) code with the text pass
        a tuple:
            (text, lang)
        Language codes are 3 *bytes* of ascii data.
        """
        if isinstance(tos, tuple):
            tos, lang = tos
        else:
            lang = DEFAULT_LANG
        if self.frame_set[frames.TOS_FID]:
            self.frame_set[frames.TOS_FID][0].text = tos
            self.frame_set[frames.TOS_FID][0].lang = lang
        else:
            self.frame_set[frames.TOS_FID] = frames.TermsOfUseFrame(text=tos, lang=lang)

    def _setCopyright(self, copyrt):
        self.setTextFrame(frames.COPYRIGHT_FID, copyrt)

    def _getCopyright(self):
        if frames.COPYRIGHT_FID in self.frame_set:
            return self.frame_set[frames.COPYRIGHT_FID][0].text

    copyright = property(_getCopyright, _setCopyright)

    def _setEncodedBy(self, enc):
        self.setTextFrame(frames.ENCODED_BY_FID, enc)

    def _getEncodedBy(self):
        if frames.ENCODED_BY_FID in self.frame_set:
            return self.frame_set[frames.ENCODED_BY_FID][0].text

    encoded_by = property(_getEncodedBy, _setEncodedBy)

    def _raiseIfReadonly(self):
        if self.read_only:
            raise RuntimeError("Tag is set read only.")

    def save(self, filename=None, version=None, encoding=None, backup=False,
             preserve_file_time=False, max_padding=None):
        """Save the tag. If ``filename`` is not give the value from the
        ``file_info`` member is used, or a ``TagException`` is raised. The
        ``version`` argument can be used to select an ID3 version other than
        the version read. ``Select text encoding with ``encoding`` or use
        the existing (or default) encoding. If ``backup`` is True the orignal
        file is preserved; likewise if ``preserve_file_time`` is True the
        file´s modification/access times are not updated.
        """
        self._raiseIfReadonly()

        if not (filename or self.file_info):
            raise TagException("No file")
        elif filename:
            self.file_info = FileInfo(filename)

        version = version if version else self.version
        if version == ID3_V2_2:
            raise NotImplementedError("Unable to write ID3 v2.2")
        self.version = version

        if backup and os.path.isfile(self.file_info.name):
            backup_name = "%s.%s" % (self.file_info.name, "orig")
            i = 1
            while os.path.isfile(backup_name):
                backup_name = "%s.%s.%d" % (self.file_info.name, "orig", i)
                i += 1
            shutil.copyfile(self.file_info.name, backup_name)

        if version[0] == 1:
            self._saveV1Tag(version)
        elif version[0] == 2:
            self._saveV2Tag(version, encoding, max_padding)
        else:
            assert(not "Version bug: %s" % str(version))

        if preserve_file_time and None not in (self.file_info.atime,
                                               self.file_info.mtime):
            self.file_info.touch((self.file_info.atime, self.file_info.mtime))
        else:
            self.file_info.initStatTimes()

    def _saveV1Tag(self, version):
        self._raiseIfReadonly()

        assert(version[0] == 1)

        def pack(s, n):
            assert(type(s) is bytes)
            if len(s) > n:
                log.warning(f"ID3 v1.x text value truncated to length {n}")
            return s.ljust(n, b'\x00')[:n]

        def encode(s):
            return s.encode("latin_1", "replace")

        # Build tag buffer.
        tag = b"TAG"
        tag += pack(encode(self.title) if self.title else b"", ID3_V1_MAX_TEXTLEN)
        tag += pack(encode(self.artist) if self.artist else b"", ID3_V1_MAX_TEXTLEN)
        tag += pack(encode(self.album) if self.album else b"", ID3_V1_MAX_TEXTLEN)

        release_date = self.getBestDate()
        year = str(release_date.year).encode("ascii") if release_date else b""
        tag += pack(year, 4)

        cmt = ""
        for c in self.comments:
            if c.description == ID3_V1_COMMENT_DESC:
                cmt = c.text
                # We prefer this one over ""
                break
            elif c.description == "":
                cmt = c.text
                # Keep searching in case we find the description eyeD3 uses.
        cmt = pack(encode(cmt), ID3_V1_MAX_TEXTLEN)

        if version != ID3_V1_0:
            track = self.track_num[0]
            if track is not None:
                cmt = cmt[0:28] + b"\x00" + bytes([int(track) & 0xff])
        tag += cmt

        if not self.genre or self.genre.id is None:
            genre = 12  # Other
        else:
            genre = self.genre.id
        tag += bytes([genre & 0xff])

        assert len(tag) == 128

        mode = "rb+" if os.path.isfile(self.file_info.name) else "w+b"
        with open(self.file_info.name, mode) as tag_file:
            # Write the tag over top an original or append it.
            try:
                tag_file.seek(-128, 2)
                if tag_file.read(3) == b"TAG":
                    tag_file.seek(-128, 2)
                else:
                    tag_file.seek(0, 2)
            except IOError:
                # File is smaller than 128 bytes.
                tag_file.seek(0, 2)

            tag_file.write(tag)
            tag_file.flush()

    def _checkForConversions(self, target_version):
        """Check the current frame set against `target_version` for frames
        requiring conversion.
        :param: The version the frames need to map to.
        :returns: A 2-tuple where the first element is a list of frames that
            are accepted for `target_version`, and the second a list of frames
            requiring conversion.
        """
        std_frames = []
        non_std_frames = []
        for f in self.frame_set.getAllFrames():
            try:
                _, fversion, _ = frames.ID3_FRAMES[f.id]
                if fversion in (target_version, ID3_V2):
                    std_frames.append(f)
                else:
                    non_std_frames.append(f)
            except KeyError:
                # Not a standard frame (ID3_FRAMES)
                try:
                    _, fversion, _ = frames.NONSTANDARD_ID3_FRAMES[f.id]
                    # but is it one we can handle.
                    if fversion in (target_version, ID3_V2):
                        std_frames.append(f)
                    else:
                        non_std_frames.append(f)
                except KeyError:
                    # Don't know anything about this pass it on for the error
                    # check there.
                    non_std_frames.append(f)

        return std_frames, non_std_frames

    def _render(self, version, curr_tag_size, max_padding_size):
        converted_frames = []
        std_frames, non_std_frames = self._checkForConversions(version)
        if non_std_frames:
            converted_frames = self._convertFrames(std_frames, non_std_frames,
                                                   version)

        # Render all frames first so the data size is known for the tag header.
        frame_data = b""
        for f in std_frames + converted_frames:
            frame_header = frames.FrameHeader(f.id, version)
            if f.header:
                frame_header.copyFlags(f.header)
            f.header = frame_header

            log.debug("Rendering frame: %s" % frame_header.id)
            raw_frame = f.render()
            log.debug("Rendered %d bytes" % len(raw_frame))
            frame_data += raw_frame

        log.debug("Rendered %d total frame bytes" % len(frame_data))

        # eyeD3 never writes unsync'd data
        self.header.unsync = False

        pending_size = TagHeader.SIZE + len(frame_data)
        if self.header.extended:
            # Using dummy data and padding, the actual size of this header
            # will be the same regardless, it's more about the flag bits
            tmp_ext_header_data = self.extended_header.render(version,
                                                              b"\x00", 0)
            pending_size += len(tmp_ext_header_data)

        if pending_size > curr_tag_size:
            # current tag (minus padding) larger than the current (plus padding)
            padding_size = DEFAULT_PADDING
            rewrite_required = True
        else:
            padding_size = curr_tag_size - pending_size
            if max_padding_size is not None and padding_size > max_padding_size:
                padding_size = min(DEFAULT_PADDING, max_padding_size)
                rewrite_required = True
            else:
                rewrite_required = False

        assert(padding_size >= 0)
        log.debug("Using %d bytes of padding" % padding_size)

        # Extended header
        ext_header_data = b""
        if self.header.extended:
            log.debug("Rendering extended header")
            ext_header_data += self.extended_header.render(self.header.version,
                                                           frame_data,
                                                           padding_size)

        # Render the tag header.
        total_size = pending_size + padding_size
        log.debug("Rendering %s tag header with size %d" %
                  (versionToString(version),
                   total_size - TagHeader.SIZE))
        header_data = self.header.render(total_size - TagHeader.SIZE)

        # Assemble the entire tag.
        tag_data = (header_data +
                    ext_header_data +
                    frame_data)
        assert(len(tag_data) == (total_size - padding_size))
        return rewrite_required, tag_data, b"\x00" * padding_size

    def _saveV2Tag(self, version, encoding, max_padding):
        self._raiseIfReadonly()

        assert(version[0] == 2 and version[1] != 2)

        log.debug("Rendering tag version: %s" % versionToString(version))

        file_exists = os.path.exists(self.file_info.name)

        if encoding:
            # Any invalid encoding is going to get coersed to a valid value
            # when the frame is rendered.
            for f in self.frame_set.getAllFrames():
                f.encoding = frames.stringToEncoding(encoding)

        curr_tag_size = 0

        if file_exists:
            # We may be converting from 1.x to 2.x so we need to find any
            # current v2.x tag otherwise we're gonna hork the file.
            # This also resets all offsets, state, etc. and makes me feel safe.
            tmp_tag = Tag()
            if tmp_tag.parse(self.file_info.name, ID3_V2):
                log.debug("Found current v2.x tag:")
                curr_tag_size = tmp_tag.file_info.tag_size
                log.debug("Current tag size: %d" % curr_tag_size)

            rewrite_required, tag_data, padding = self._render(version,
                                                               curr_tag_size,
                                                               max_padding)
            log.debug("Writing %d bytes of tag data and %d bytes of "
                      "padding" % (len(tag_data), len(padding)))
            if rewrite_required:
                # Open tmp file
                with tempfile.NamedTemporaryFile("wb", delete=False) \
                        as tmp_file:
                    tmp_file.write(tag_data + padding)

                    # Copy audio data in chunks
                    with open(self.file_info.name, "rb") as tag_file:
                        if curr_tag_size != 0:
                            seek_point = curr_tag_size
                        else:
                            seek_point = 0
                        log.debug("Seeking to beginning of audio data, "
                                  "byte %d (%x)" % (seek_point, seek_point))
                        tag_file.seek(seek_point)
                        chunkCopy(tag_file, tmp_file)

                    tmp_file.flush()

                # Move tmp to orig.
                shutil.copyfile(tmp_file.name, self.file_info.name)
                os.unlink(tmp_file.name)

            else:
                with open(self.file_info.name, "r+b") as tag_file:
                    tag_file.write(tag_data + padding)

        else:
            _, tag_data, padding = self._render(version, 0, None)
            with open(self.file_info.name, "wb") as tag_file:
                tag_file.write(tag_data + padding)

        log.debug("Tag write complete. Updating FileInfo state.")
        self.file_info.tag_size = len(tag_data) + len(padding)

    def _convertFrames_v1(self, std_frames, convert_list, version) -> list:
        assert version[0] == 1
        converted_frames = []

        track_num_frame = None
        for frame in std_frames:
            if frame.id == frames.TRACKNUM_FID:
                # Find track_num so it can be enforced for 1.1
                track_num_frame = frame
            elif frame.id == frames.COMMENT_FID and frame.description == ID3_V1_COMMENT_DESC:
                # Comments truncated to make room for v1.1 track
                if version == ID3_V1_1:
                    if len(frame.text) > ID3_V1_MAX_TEXTLEN - 2:
                        trunc_text = frame.text[:ID3_V1_MAX_TEXTLEN - 2]
                        log.info(f"Truncating ID3 v1 comment due to tag conversion: {frame.text}")
                        frame.text = trunc_text

        # v1.1 must have a track num
        if track_num_frame is None and version == ID3_V1_1:
            log.info("ID3 v1.0->v1.1 conversion forces track number, defaulting to 1")
            std_frames.append(frames.TextFrame(frames.TRACKNUM_FID, "1"))
        # v1.0 must not
        elif track_num_frame is not None and version == ID3_V1_0:
            log.info("ID3 v1.1->v1.0 conversion forces deleting track number")
            std_frames.remove(track_num_frame)

        for frame in list(convert_list):
            # Let date frames thru, the right thing will happen on save
            if isinstance(frame, frames.DateFrame):
                converted_frames.append(frame)
                convert_list.remove(frame)

        return converted_frames

    def _convertFrames(self, std_frames, convert_list, version) -> list:
        """Maps frame incompatibilities between ID3 tag versions.

        The items in ``std_frames`` need no conversion, but the list/frames
        may be edited if necessary (e.g. a converted frame replaces a frame
        in the list).  The items in ``convert_list`` are the frames to convert
        and return. The ``version`` is the target ID3 version."""
        from . import versionToString
        from .frames import DATE_FIDS, DEPRECATED_DATE_FIDS, DateFrame, TextFrame

        if version[0] == 1:
            return self._convertFrames_v1(std_frames, convert_list, version)

        # Only ID3 v2.x onward
        assert version[0] != 1
        converted_frames = []
        flist = list(convert_list)

        # Date frame conversions.
        date_frames = {}
        for f in flist:
            if version == ID3_V2_4:
                if f.id in DEPRECATED_DATE_FIDS:
                    date_frames[f.id] = f
            else:
                if f.id in DATE_FIDS:
                    date_frames[f.id] = f

        if date_frames:
            def fidHandled(_fid):
                # A duplicate text frame (illegal ID3 but oft seen) may exist. The date_frames dict
                # will have one, but the flist has multiple, hence the loop.
                for _frame in list(flist):
                    if _frame.id == _fid:
                        flist.remove(_frame)
                del date_frames[_fid]

            if version == ID3_V2_4:
                if b"TORY" in date_frames or b"XDOR" in date_frames:
                    # XDOR -> TDOR (full date)
                    # TORY -> TDOR (year only)
                    date = self._getV23OriginalReleaseDate()
                    if date:
                        converted_frames.append(DateFrame(b"TDOR", date))
                    for fid in (b"TORY", b"XDOR"):
                        if fid in flist:
                            fidHandled(fid)

                # TYER, TDAT, TIME -> TDRC
                if (b"TYER" in date_frames or b"TDAT" in date_frames or b"TIME" in date_frames):
                    date = self._getV23RecordingDate()
                    if date:
                        converted_frames.append(DateFrame(b"TDRC", date))
                    for fid in [b"TYER", b"TDAT", b"TIME"]:
                        if fid in date_frames:
                            fidHandled(fid)

            elif version == ID3_V2_3:
                if b"TDOR" in date_frames:
                    date = date_frames[b"TDOR"].date
                    if date:
                        # TORY is year only
                        converted_frames.append(DateFrame(b"TORY", str(date.year)))
                    if date and date.month:
                        converted_frames.append(DateFrame(b"XDOR", str(date)))
                    fidHandled(b"TDOR")

                if b"TDRC" in date_frames:
                    date = date_frames[b"TDRC"].date

                    if date:
                        converted_frames.append(DateFrame(b"TYER", str(date.year)))
                        if None not in (date.month, date.day):
                            date_str = "%s%s" %\
                                    (str(date.day).rjust(2, "0"),
                                     str(date.month).rjust(2, "0"))
                            converted_frames.append(TextFrame(b"TDAT",
                                                              date_str))
                        if None not in (date.hour, date.minute):
                            date_str = "%s%s" %\
                                    (str(date.hour).rjust(2, "0"),
                                     str(date.minute).rjust(2, "0"))
                            converted_frames.append(TextFrame(b"TIME",
                                                              date_str))

                    fidHandled(b"TDRC")

                if b"TDRL" in date_frames:
                    # TDRL -> Nothing
                    log.warning("TDRL value dropped.")
                    fidHandled(b"TDRL")

            # All other date frames have no conversion
            for fid in date_frames:
                log.warning(f"{str(fid, 'ascii')} frame being dropped due to conversion to "
                            f"{versionToString(version)}")
                flist.remove(date_frames[fid])

        # Convert sort order frames 2.3 (XSO*) <-> 2.4 (TSO*)
        prefix = b"X" if version == ID3_V2_4 else b"T"
        fids = [prefix + suffix for suffix in [b"SOA", b"SOP", b"SOT"]]
        soframes = [f for f in flist if f.id in fids]

        for frame in soframes:
            frame.id = (b"X" if prefix == b"T" else b"T") + frame.id[1:]
            flist.remove(frame)
            converted_frames.append(frame)

        # TSIZ (v2.3) are completely deprecated, remove them
        if version == ID3_V2_4:
            flist = [f for f in flist if f.id != b"TSIZ"]

        # TSST (v2.4) --> TIT3 (2.3)
        if version == ID3_V2_3 and b"TSST" in [f.id for f in flist]:
            tsst_frame = [f for f in flist if f.id == b"TSST"][0]
            flist.remove(tsst_frame)
            tsst_frame = frames.UserTextFrame(
                    description="Subtitle (converted)", text=tsst_frame.text)
            converted_frames.append(tsst_frame)

        # RVAD (v2.3) --> RVA2* (2.4)
        if version == ID3_V2_4 and b"RVAD" in [f.id for f in flist]:
            rvad = [f for f in flist if f.id == b"RVAD"][0]
            for rva2 in rvad.toV24():
                converted_frames.append(rva2)
            flist.remove(rvad)
        # RVA2* (v2.4) --> RVAD (2.3)
        elif version == ID3_V2_3 and b"RVA2" in [f.id for f in flist]:
            adj = frames.RelVolAdjFrameV23.VolumeAdjustments()
            for rva2 in [f for f in flist if f.id == b"RVA2"]:
                adj.setChannelAdj(rva2.channel_type, rva2.adjustment * 512)
                adj.setChannelPeak(rva2.channel_type, rva2.peak)
                flist.remove(rva2)

            rvad = frames.RelVolAdjFrameV23()
            rvad.adjustments = adj
            converted_frames.append(rvad)

        # Raise an error for frames that could not be converted.
        if len(flist) != 0:
            unconverted = ", ".join([f.id.decode("ascii") for f in flist])
            if version[0] != 1:
                raise TagException("Unable to convert the following frames to "
                                   f"version {versionToString(version)}: {unconverted}")

        # Some frames in converted_frames may replace/edit frames in std_frames.
        for cframe in converted_frames:
            for sframe in std_frames:
                if cframe.id == sframe.id:
                    std_frames.remove(sframe)

        return converted_frames

    @staticmethod
    def remove(filename, version=ID3_ANY_VERSION, preserve_file_time=False):
        tag = None
        retval = False

        if version[0] & ID3_V1[0]:
            # ID3 v1.x
            tag = Tag()
            with open(filename, "r+b") as tag_file:
                found = tag.parse(tag_file, ID3_V1)
                if found:
                    tag_file.seek(-128, 2)
                    log.debug("Removing ID3 v1.x Tag")
                    tag_file.truncate()
                    retval |= True

        if version[0] & ID3_V2[0]:
            tag = Tag()
            with open(filename, "rb") as tag_file:
                found = tag.parse(tag_file, ID3_V2)
                if found:
                    log.debug("Removing ID3 %s tag" %
                              versionToString(tag.version))
                    tag_file.seek(tag.file_info.tag_size)

                    # Open tmp file
                    with tempfile.NamedTemporaryFile("wb", delete=False) \
                            as tmp_file:
                        chunkCopy(tag_file, tmp_file)

                    # Move tmp to orig
                    shutil.copyfile(tmp_file.name, filename)
                    os.unlink(tmp_file.name)

                    retval |= True

        if preserve_file_time and retval and None not in (tag.file_info.atime,
                                                          tag.file_info.mtime):
            tag.file_info.touch((tag.file_info.atime, tag.file_info.mtime))

        return retval

    @property
    def chapters(self):
        return self._chapters

    @property
    def table_of_contents(self):
        return self._tocs

    @property
    def album_type(self):
        if TXXX_ALBUM_TYPE in self.user_text_frames:
            return self.user_text_frames.get(TXXX_ALBUM_TYPE).text
        else:
            return None

    @album_type.setter
    def album_type(self, t):
        if not t:
            self.user_text_frames.remove(TXXX_ALBUM_TYPE)
        elif t in ALBUM_TYPE_IDS:
            self.user_text_frames.set(t, TXXX_ALBUM_TYPE)
        else:
            raise ValueError("Invalid album_type: %s" % t)

    @property
    def artist_origin(self):
        """Returns None or a `ArtistOrigin` dataclass: (city, state, country) Any may be ``None``.
        """
        if TXXX_ARTIST_ORIGIN not in self.user_text_frames:
            return None

        origin = self.user_text_frames.get(TXXX_ARTIST_ORIGIN).text
        vals = origin.split('\t')

        vals.extend([None] * (3 - len(vals)))
        vals = [None if not v else v for v in vals]
        return ArtistOrigin(*vals)

    @artist_origin.setter
    def artist_origin(self, origin: ArtistOrigin):
        if origin is None or origin == (None, None, None):
            self.user_text_frames.remove(TXXX_ARTIST_ORIGIN)
        else:
            self.user_text_frames.set(origin.id3Encode(), TXXX_ARTIST_ORIGIN)

    def frameiter(self, fids=None):
        """A iterator for tag frames. If ``fids`` is passed it must be a list
        of frame IDs to filter and return."""
        fids = fids or []
        fids = [(b(f, ascii_encode) if isinstance(f, str) else f) for f in fids]
        for f in self.frame_set.getAllFrames():
            if not fids or f.id in fids:
                yield f

    def _getOrigArtist(self):
        return self.getTextFrame(frames.ORIG_ARTIST_FID)

    def _setOrigArtist(self, name):
        self.setTextFrame(frames.ORIG_ARTIST_FID, name)

    @property
    def original_artist(self):
        return self._getOrigArtist()

    @original_artist.setter
    def original_artist(self, name):
        self._setOrigArtist(name)


class FileInfo:
    """
    This class is for storing information about a parsed file. It contains info
    such as the filename, original tag size, and amount of padding; all of which
    can make rewriting faster.
    """
    def __init__(self, file_name, tagsz=0, tpadd=0):
        from .. import LOCAL_FS_ENCODING

        if type(file_name) is str:
            self.name = file_name
        else:
            try:
                self.name = str(file_name, LOCAL_FS_ENCODING)
            except UnicodeDecodeError:
                # Work around the local encoding not matching that of a mounted
                # filesystem
                log.warning("Mismatched file system encoding for file '%s'" %
                            repr(file_name))
                self.name = file_name

        self.tag_size = tagsz or 0  # This includes the padding byte count.
        self.tag_padding_size = tpadd or 0

        self.atime, self.mtime = None, None
        self.initStatTimes()

    def initStatTimes(self):
        try:
            s = os.stat(self.name)
        except OSError:
            self.atime, self.mtime = None, None
        else:
            self.atime, self.mtime = s.st_atime, s.st_mtime

    def touch(self, times):
        """times is a 2-tuple of (atime, mtime)."""
        os.utime(self.name, times)
        self.initStatTimes()


class AccessorBase:
    def __init__(self, fid, fs, match_func=None):
        self._fid = fid
        self._fs = fs
        self._match_func = match_func

    def __iter__(self):
        for f in self._fs[self._fid] or []:
            yield f

    def __len__(self):
        return len(self._fs[self._fid] or [])

    def __getitem__(self, i):
        frames = self._fs[self._fid]
        if not frames:
            raise IndexError("list index out of range")
        return frames[i]

    def get(self, *args, **kwargs):
        for frame in self._fs[self._fid] or []:
            if self._match_func(frame, *args, **kwargs):
                return frame
        return None

    def remove(self, *args, **kwargs):
        """Returns the removed item or ``None`` if not found."""
        fid_frames = self._fs[self._fid] or []
        for frame in fid_frames:
            if self._match_func(frame, *args, **kwargs):
                fid_frames.remove(frame)
                return frame
        return None


class DltAccessor(AccessorBase):
    """Access matching tag frames by "description" and/or "lang" values."""
    def __init__(self, FrameClass, fid, fs):
        def match_func(frame, description, lang=DEFAULT_LANG):
            return (frame.description == description and
                    frame.lang == (lang if isinstance(lang, bytes)
                                        else lang.encode("ascii")))

        super().__init__(fid, fs, match_func)
        self.FrameClass = FrameClass

    @requireUnicode(1, 2)
    def set(self, text, description="", lang=DEFAULT_LANG):
        lang = lang or DEFAULT_LANG
        for f in self._fs[self._fid] or []:
            if f.description == description and f.lang == lang:
                # Exists, update text
                f.text = text
                return f

        new_frame = self.FrameClass(description=description, lang=lang,
                                    text=text)
        self._fs[self._fid] = new_frame
        return new_frame

    @requireUnicode(1)
    def remove(self, description, lang=DEFAULT_LANG):
        return super().remove(description, lang=lang or DEFAULT_LANG)

    @requireUnicode(1)
    def get(self, description, lang=DEFAULT_LANG):
        return super().get(description, lang=lang or DEFAULT_LANG)


class CommentsAccessor(DltAccessor):
    def __init__(self, fs):
        super().__init__(frames.CommentFrame, frames.COMMENT_FID, fs)


class LyricsAccessor(DltAccessor):
    def __init__(self, fs):
        super().__init__(frames.LyricsFrame, frames.LYRICS_FID, fs)


class ImagesAccessor(AccessorBase):
    def __init__(self, fs):
        def match_func(frame, description):
            return frame.description == description
        super().__init__(frames.IMAGE_FID, fs, match_func)

    @requireUnicode("description")
    def set(self, type_, img_data, mime_type, description="", img_url=None):
        """Add an image of ``type_`` (a type constant from ImageFrame).
        The ``img_data`` is either bytes or ``None``. In the latter case
        ``img_url`` MUST be the URL to the image. In this case ``mime_type``
        is ignored and "-->" is used to signal this as a link and not data
        (per the ID3 spec)."""
        img_url = b(img_url) if img_url else None

        if not img_data and not img_url:
            raise ValueError("img_url MUST not be none when no image data")

        mime_type = mime_type if img_data else frames.ImageFrame.URL_MIME_TYPE
        mime_type = b(mime_type)

        images = self._fs[frames.IMAGE_FID] or []
        for img in images:
            if img.description == description:
                # update
                if not img_data:
                    img.image_url = img_url
                    img.image_data = None
                    img.mime_type = frames.ImageFrame.URL_MIME_TYPE
                else:
                    img.image_url = None
                    img.image_data = img_data
                    img.mime_type = mime_type
                img.picture_type = type_
                return img

        img_frame = frames.ImageFrame(description=description,
                                      image_data=img_data,
                                      image_url=img_url,
                                      mime_type=mime_type,
                                      picture_type=type_)
        self._fs[frames.IMAGE_FID] = img_frame
        return img_frame

    @requireUnicode(1)
    def remove(self, description):
        return super().remove(description)

    @requireUnicode(1)
    def get(self, description):
        return super().get(description)


class ObjectsAccessor(AccessorBase):
    def __init__(self, fs):
        def match_func(frame, description):
            return frame.description == description
        super().__init__(frames.OBJECT_FID, fs, match_func)

    @requireUnicode("description", "filename")
    def set(self, data, mime_type, description="", filename=""):
        objects = self._fs[frames.OBJECT_FID] or []
        for obj in objects:
            if obj.description == description:
                # update
                obj.object_data = data
                obj.mime_type = mime_type
                obj.filename = filename
                return obj

        obj_frame = frames.ObjectFrame(description=description,
                                       filename=filename,
                                       object_data=data,
                                       mime_type=mime_type)
        self._fs[frames.OBJECT_FID] = obj_frame
        return obj_frame

    @requireUnicode(1)
    def remove(self, description):
        return super().remove(description)

    @requireUnicode(1)
    def get(self, description):
        return super().get(description)


class PrivatesAccessor(AccessorBase):
    def __init__(self, fs):
        def match_func(frame, owner_id):
            return frame.owner_id == owner_id
        super().__init__(frames.PRIVATE_FID, fs, match_func)

    def set(self, data, owner_id):
        priv_frames = self._fs[frames.PRIVATE_FID] or []
        for f in priv_frames:
            if f.owner_id == owner_id:
                # update
                f.owner_data = data
                return f

        priv_frame = frames.PrivateFrame(owner_id=owner_id,
                                         owner_data=data)
        self._fs[frames.PRIVATE_FID] = priv_frame
        return priv_frame

    def remove(self, owner_id):
        return super().remove(owner_id)

    def get(self, owner_id):
        return super().get(owner_id)


class UserTextsAccessor(AccessorBase):
    def __init__(self, fs):
        def match_func(frame, description):
            return frame.description == description
        super().__init__(frames.USERTEXT_FID, fs, match_func)

    @requireUnicode(1, "description")
    def set(self, text, description=""):
        flist = self._fs[frames.USERTEXT_FID] or []
        for utf in flist:
            if utf.description == description:
                # update
                utf.text = text
                return utf

        utf = frames.UserTextFrame(description=description,
                                   text=text)
        self._fs[frames.USERTEXT_FID] = utf
        return utf

    @requireUnicode(1)
    def remove(self, description):
        return super().remove(description)

    @requireUnicode(1)
    def get(self, description):
        return super().get(description)

    @requireUnicode(1)
    def __contains__(self, description):
        return bool(self.get(description))


class UniqueFileIdAccessor(AccessorBase):
    def __init__(self, fs):
        def match_func(frame, owner_id):
            return frame.owner_id == owner_id
        super().__init__(frames.UNIQUE_FILE_ID_FID, fs, match_func)

    def set(self, data, owner_id):
        data, owner_id = b(data), b(owner_id)
        if len(data) > 64:
            raise TagException("UFID data must be 64 bytes or less")

        flist = self._fs[frames.UNIQUE_FILE_ID_FID] or []
        for f in flist:
            if f.owner_id == owner_id:
                # update
                f.uniq_id = data
                return f

        uniq_id_frame = frames.UniqueFileIDFrame(owner_id=owner_id,
                                                 uniq_id=data)
        self._fs[frames.UNIQUE_FILE_ID_FID] = uniq_id_frame
        return uniq_id_frame

    def remove(self, owner_id):
        owner_id = b(owner_id)
        return super().remove(owner_id)

    def get(self, owner_id):
        owner_id = b(owner_id)
        return super().get(owner_id)


class UserUrlsAccessor(AccessorBase):
    def __init__(self, fs):
        def match_func(frame, description):
            return frame.description == description
        super().__init__(frames.USERURL_FID, fs, match_func)

    @requireUnicode("description")
    def set(self, url, description=""):
        flist = self._fs[frames.USERURL_FID] or []
        for uuf in flist:
            if uuf.description == description:
                # update
                uuf.url = url
                return uuf

        uuf = frames.UserUrlFrame(description=description, url=url)
        self._fs[frames.USERURL_FID] = uuf
        return uuf

    @requireUnicode(1)
    def remove(self, description):
        return super().remove(description)

    @requireUnicode(1)
    def get(self, description):
        return super().get(description)


class PopularitiesAccessor(AccessorBase):
    def __init__(self, fs):
        def match_func(frame, email):
            return frame.email == email
        super().__init__(frames.POPULARITY_FID, fs, match_func)

    def set(self, email, rating, play_count):
        flist = self._fs[frames.POPULARITY_FID] or []
        for popm in flist:
            if popm.email == email:
                # update
                popm.rating = rating
                popm.count = play_count
                return popm

        popm = frames.PopularityFrame(email=email, rating=rating,
                                      count=play_count)
        self._fs[frames.POPULARITY_FID] = popm
        return popm

    def remove(self, email):
        return super().remove(email)

    def get(self, email):
        return super().get(email)


class ChaptersAccessor(AccessorBase):
    def __init__(self, fs):
        def match_func(frame, element_id):
            return frame.element_id == element_id
        super().__init__(frames.CHAPTER_FID, fs, match_func)

    def set(self, element_id, times, offsets=(None, None), sub_frames=None):
        flist = self._fs[frames.CHAPTER_FID] or []
        for chap in flist:
            if chap.element_id == element_id:
                # update
                chap.times, chap.offsets = times, offsets
                if sub_frames:
                    chap.sub_frames = sub_frames
                return chap

        chap = frames.ChapterFrame(element_id=element_id,
                                   times=times, offsets=offsets,
                                   sub_frames=sub_frames)
        self._fs[frames.CHAPTER_FID] = chap
        return chap

    def remove(self, element_id):
        return super().remove(element_id)

    def get(self, element_id):
        return super().get(element_id)

    def __getitem__(self, elem_id):
        """Overiding the index based __getitem__ for one indexed with chapter
        element IDs. These are stored in the tag's table of contents frames."""
        for chapter in (self._fs[frames.CHAPTER_FID] or []):
            if chapter.element_id == elem_id:
                return chapter
        raise IndexError("chapter '%s' not found" % elem_id)


class TocAccessor(AccessorBase):
    def __init__(self, fs):
        def match_func(frame, element_id):
            return frame.element_id == element_id
        super().__init__(frames.TOC_FID, fs, match_func)

    def __iter__(self):
        tocs = list(self._fs[self._fid] or [])
        for toc_frame in tocs:
            # Find and put top level at the front of the list
            if toc_frame.toplevel:
                tocs.remove(toc_frame)
                tocs.insert(0, toc_frame)
                break

        for toc in tocs:
            yield toc

    @requireUnicode("description")
    def set(self, element_id, toplevel=False, ordered=True, child_ids=None,
            description=""):
        flist = self._fs[frames.TOC_FID] or []

        # Enforce one top-level
        if toplevel:
            for toc in flist:
                if toc.toplevel:
                    raise ValueError("There may only be one top-level "
                                     "table of contents. Toc '%s' is current "
                                     "top-level." % toc.element_id)
        for toc in flist:
            if toc.element_id == element_id:
                # update
                toc.toplevel = toplevel
                toc.ordered = ordered
                toc.child_ids = child_ids
                toc.description = description
                return toc

        toc = frames.TocFrame(element_id=element_id, toplevel=toplevel,
                              ordered=ordered, child_ids=child_ids,
                              description=description)
        self._fs[frames.TOC_FID] = toc
        return toc

    def remove(self, element_id):
        return super().remove(element_id)

    def get(self, element_id):
        return super().get(element_id)

    def __getitem__(self, elem_id):
        """Overiding the index based __getitem__ for one indexed with table
        of contents element IDs."""
        for toc in (self._fs[frames.TOC_FID] or []):
            if toc.element_id == elem_id:
                return toc
        raise IndexError("toc '%s' not found" % elem_id)


class TagTemplate(string.Template):
    idpattern = r'[_a-z][_a-z0-9:]*'

    def __init__(self, pattern, path_friendly="-", dotted_dates=False):
        super().__init__(pattern)

        if type(path_friendly) is bool and path_friendly:
            # Previous versions used boolean values, convert old default to new
            path_friendly = "-"
        self._path_friendly = path_friendly

        self._dotted_dates = dotted_dates

    def substitute(self, tag, zeropad=True):
        mapping = self._makeMapping(tag, zeropad)

        # Helper function for .sub()
        def convert(mo):
            named = mo.group('named')
            if named is not None:
                try:
                    if type(mapping[named]) is tuple:
                        func, args = mapping[named][0], mapping[named][1:]
                        return '%s' % func(tag, named, *args)
                    # We use this idiom instead of str() because the latter
                    # will fail if val is a Unicode containing non-ASCII
                    return '%s' % (mapping[named],)
                except KeyError:
                    return self.delimiter + named
            braced = mo.group('braced')
            if braced is not None:
                try:
                    if type(mapping[braced]) is tuple:
                        func, args = mapping[braced][0], mapping[braced][1:]
                        return '%s' % func(tag, braced, *args)
                    return '%s' % (mapping[braced],)
                except KeyError:
                    return self.delimiter + '{' + braced + '}'
            if mo.group('escaped') is not None:
                return self.delimiter
            if mo.group('invalid') is not None:
                return self.delimiter
            raise ValueError('Unrecognized named group in pattern',
                             self.pattern)

        name = self.pattern.sub(convert, self.template)
        if self._path_friendly:
            name = name.replace("/", self._path_friendly)
        return name

    safe_substitute = substitute

    def _dates(self, tag, param):
        if param.startswith("release_"):
            date = tag.release_date
        elif param.startswith("recording_"):
            date = tag.recording_date
        elif param.startswith("original_release_"):
            date = tag.original_release_date
        else:
            date = tag.getBestDate(
                    prefer_recording_date=":prefer_recording" in param)

        if date and param.endswith(":year"):
            dstr = str(date.year)
        elif date:
            dstr = str(date)
        else:
            dstr = ""

        if self._dotted_dates:
            dstr = dstr.replace('-', '.')

        return dstr

    @staticmethod
    def _nums(num_tuple, param, zeropad):
        nn, nt = ((str(n) if n else None) for n in num_tuple)
        if zeropad:
            if nt:
                nt = nt.rjust(2, "0")
            nn = nn.rjust(len(nt) if nt else 2, "0")

        if param.endswith(":num"):
            return nn
        elif param.endswith(":total"):
            return nt
        else:
            raise ValueError("Unknown template param: %s" % param)

    def _track(self, tag, param, zeropad):
        return self._nums(tag.track_num, param, zeropad)

    def _disc(self, tag, param, zeropad):
        return self._nums(tag.disc_num, param, zeropad)

    @staticmethod
    def _file(tag, param):
        assert(param.startswith("file"))

        if param.endswith(":ext"):
            return os.path.splitext(tag.file_info.name)[1][1:]
        else:
            return tag.file_info.name

    def _makeMapping(self, tag, zeropad):
        return {"artist": tag.artist if tag else None,
                "album_artist": tag.album_artist if tag else None,
                "album": tag.album if tag else None,
                "title": tag.title if tag else None,
                "track:num": (self._track, zeropad) if tag else None,
                "track:total": (self._track, zeropad) if tag else None,
                "release_date": (self._dates,) if tag else None,
                "release_date:year": (self._dates,) if tag else None,
                "recording_date": (self._dates,) if tag else None,
                "recording_date:year": (self._dates,) if tag else None,
                "original_release_date": (self._dates,) if tag else None,
                "original_release_date:year": (self._dates,) if tag else None,
                "best_date": (self._dates,) if tag else None,
                "best_date:year": (self._dates,) if tag else None,
                "best_date:prefer_recording": (self._dates,) if tag else None,
                "best_date:prefer_release": (self._dates,) if tag else None,
                "best_date:prefer_recording:year": (self._dates,) if tag
                                                                  else None,
                "best_date:prefer_release:year": (self._dates,) if tag
                                                                   else None,
                "file": (self._file,) if tag else None,
                "file:ext": (self._file,) if tag else None,
                "disc:num": (self._disc, zeropad) if tag else None,
                "disc:total": (self._disc, zeropad) if tag else None,
               }
